# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs planarity
"""
Schnyder's algorithm for straight-line planar embeddings

A module for computing the (x,y) coordinates for a straight-line planar
embedding of any connected planar graph with at least three vertices.  Uses
Walter Schnyder's Algorithm from [Sch1990]_.

AUTHORS:

- Jonathan Bober, Emily Kirkman (2008-02-09) --  initial version
"""
# ****************************************************************************
#      Copyright (C) 2008 Jonathan Bober and Emily Kirkman
#
# Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
#                         https://www.gnu.org/licenses/
# ****************************************************************************

from sage.sets.set import Set
from .all import DiGraph


def _triangulate(g, comb_emb):
    """
    Helper function to schnyder method for computing coordinates in the plane to
    plot a planar graph with no edge crossings.

    Given a connected graph g with at least 3 vertices and a planar combinatorial
    embedding comb_emb of g, modify g in place to form a graph whose faces are
    all triangles, and return the set of newly created edges. Also ``comb_emb``
    is updated in place.

    The simple way to triangulate a face is to just pick a vertex and draw
    an edge from that vertex to every other vertex in the face. Think that this
    might ultimately result in graphs that don't look very nice when we draw them
    so we have decided on a different strategy.  After handling special cases,
    we add an edge to connect every third vertex on each face.  If this edge is
    a repeat, we keep the first edge of the face and retry the process at the next
    edge in the face list.  (By removing the special cases we guarantee that this
    method will work on one of these attempts.)

    INPUT:

    - ``g`` -- the graph to triangulate
    - ``comb_emb`` -- a planar combinatorial embedding of g

    OUTPUT: a list of edges that are added to the graph (in place)

    EXAMPLES::

        sage: from sage.graphs.schnyder import _triangulate
        sage: g = Graph(graphs.CycleGraph(4))
        sage: g.is_planar(set_embedding=True)
        True
        sage: _triangulate(g, g._embedding)
        [(2, 0), (1, 3)]

        sage: g = graphs.PathGraph(3)
        sage: g.is_planar(set_embedding=True)
        True
        sage: new_edges = _triangulate(g, g._embedding)
        sage: [sorted(e) for e in new_edges]
        [[0, 2]]

    TESTS:

    :issue:`29522` is fixed::

        sage: g = Graph(2)
        sage: _triangulate(g, {})
        Traceback (most recent call last):
        ...
        NotImplementedError: _triangulate() only knows how to handle connected graphs
        sage: g = Graph([(0, 1)])
        sage: _triangulate(g, {})
        Traceback (most recent call last):
        ...
        ValueError: a Graph with less than 3 vertices doesn't have any triangulation
        sage: g = Graph(3)
        sage: _triangulate(g, {})
        Traceback (most recent call last):
        ...
        NotImplementedError: _triangulate() only knows how to handle connected graphs
    """
    # first make sure that the graph has at least 3 vertices, and that it is connected
    if not g.is_connected():
        raise NotImplementedError("_triangulate() only knows how to handle connected graphs")
    if g.order() < 3:
        raise ValueError("a Graph with less than 3 vertices doesn't have any triangulation")

    # At this point we know that the graph is connected, has at least 3
    # vertices. This is where the real work starts.

    faces = g.faces(comb_emb)
    # We start by finding all of the faces of this embedding.

    edges_added = []   # The list of edges that we add to the graph.
    # This will be returned at the end.

    for face in faces:
        new_face = []
        if len(face) < 3:
            raise RuntimeError('Triangulate method created face %s with < 3 edges.' % face)
        if len(face) == 3:
            continue  # This face is already triangulated
        elif len(face) == 4:  # In this special case just add diagonal edge to square
            u, v, w, x = (e[0] for e in face)
            if w == u or g.has_edge(w, u):
                u, v, w, x = v, w, x, u
            new_face = (w, u)
            comb_emb[w].insert(comb_emb[w].index(x), u)
            comb_emb[u].insert(comb_emb[u].index(v), w)
            g.add_edge(new_face)
            edges_added.append(new_face)
        else:
            N = len(face)
            i = 0
            while i < N - 1:
                new_edge = (face[i + 1][1], face[i][0])  # new_edge is from third vertex in face to first
                if g.has_edge(new_edge) or new_edge[0] == new_edge[1]:  # check for repeats
                    new_face.append(face[i])  # if repeated, keep first edge in face instead
                    if i == N - 2:   # if we are two from the end, found a triangle already
                        break
                    i += 1
                    continue

                g.add_edge(new_edge)
                edges_added.append(new_edge)
                comb_emb[new_edge[0]].insert(comb_emb[new_edge[0]].index((face + new_face)[i + 2][1]), new_edge[1])
                comb_emb[new_edge[1]].insert(comb_emb[new_edge[1]].index(face[i][1]), new_edge[0])
                new_face.append((new_edge[1], new_edge[0]))
                i += 2
            if i != N:
                new_face.append(face[-1])
            faces.append(new_face)

    return edges_added


def _normal_label(g, comb_emb, external_face):
    r"""
    Helper function to schnyder method for computing coordinates in
    the plane to plot a planar graph with no edge crossings.

    Constructs a normal labelling of a triangular graph g, given the
    planar combinatorial embedding of g and a designated external
    face.  Returns labels dictionary.  The normal label is constructed
    by first contracting the graph down to its external face, then
    expanding the graph back to the original while simultaneously
    adding angle labels.

    INPUT:

    - ``g`` -- the graph to find the normal labeling of (g must be triangulated)
    - ``comb_emb`` -- a planar combinatorial embedding of g
    - ``external_face`` -- the list of three edges in the external face of g

    OUTPUT: x; tuple with entries

        x[0] = dict of dicts of normal labeling for each vertex of g and each
        adjacent neighbors u,v (u < v) of vertex:

        { vertex : { (u,v): angel_label } }

        x[1] = (v1,v2,v3) tuple of the three vertices of the external face.

    EXAMPLES::

        sage: from sage.graphs.schnyder import _triangulate, _normal_label, _realizer
        sage: g = Graph(graphs.CycleGraph(7))
        sage: g.is_planar(set_embedding=True)
        True
        sage: faces = g.faces(g._embedding)
        sage: _triangulate(g, g._embedding)
        [(2, 0), (4, 2), (6, 4), (5, 0), (3, 5), (1, 3), (4, 0), (3, 0)]
        sage: tn = _normal_label(g, g._embedding, faces[0])
        sage: _realizer(g, tn)
        ({0: [<sage.graphs.schnyder.TreeNode object at ...>]},
         (1, 0, 2))
    """
    contracted = []
    contractible = []

    labels = {}

    # For now we will not take the order of the outer face into account.
    # We will correct this in the end of this function.
    external_vertices = sorted([external_face[0][0],
                                external_face[1][0],
                                external_face[2][0]])
    v1, v2, v3 = external_vertices
    v1_neighbors = Set(g.neighbors(v1))

    neighbor_count = {v: 0 for v in g}
    for v in g.neighbors(v1):
        neighbor_count[v] = len(v1_neighbors.intersection(Set(g.neighbors(v))))

    for v in v1_neighbors:
        if v in [v1, v2, v3]:
            continue
        if neighbor_count[v] == 2:
            contractible.append(v)

    # contraction phase:

    while g.order() > 3:
        try:
            v = contractible.pop()
        except Exception:
            raise RuntimeError('Contractible list is empty but graph still has %d vertices.  (Expected 3.)' % g.order())

        # going to contract v
        v_neighbors = Set(g.neighbors(v))
        contracted.append((v, v_neighbors,
                           v_neighbors - v1_neighbors - Set([v1])))
        g.delete_vertex(v)
        v1_neighbors -= Set([v])
        for w in v_neighbors - v1_neighbors - Set([v1]):
            # adding edge (v1, w)
            g.add_edge((v1, w))
        if g.order() == 3:
            break
        v1_neighbors += v_neighbors - Set([v1])
        contractible = []
        for w in g.neighbors(v1):
            if (len(v1_neighbors.intersection(Set(g.neighbors(w)))) == 2
                    and w not in [v1, v2, v3]):
                contractible.append(w)

    # expansion phase:

    v1, v2, v3 = g.vertices(sort=True)  # always in sorted order

    labels[v1] = {(v2, v3): 1}
    labels[v2] = {(v1, v3): 2}
    labels[v3] = {(v1, v2): 3}

    while contracted:
        v, new_neighbors, neighbors_to_delete = contracted.pop()
        # going to add back vertex v
        labels[v] = {}

        for w in neighbors_to_delete:
            g.delete_edge((v1, w))

        if len(neighbors_to_delete) == 0:
            # we are adding v into the face new_neighbors
            w1, w2, w3 = sorted(new_neighbors)

            labels[v] = {(w1, w2): labels[w3].pop((w1, w2)),
                         (w2, w3): labels[w1].pop((w2, w3)),
                         (w1, w3): labels[w2].pop((w1, w3))}
            labels[w1][tuple(sorted((w2, v)))] = labels[v][(w2, w3)]
            labels[w1][tuple(sorted((w3, v)))] = labels[v][(w2, w3)]

            labels[w2][tuple(sorted((w1, v)))] = labels[v][(w1, w3)]
            labels[w2][tuple(sorted((w3, v)))] = labels[v][(w1, w3)]

            labels[w3][tuple(sorted((w1, v)))] = labels[v][(w1, w2)]
            labels[w3][tuple(sorted((w2, v)))] = labels[v][(w1, w2)]
        else:
            new_neighbors_set = Set(new_neighbors)
            angles_out_of_v1 = set()
            vertices_in_order = []
            l = []
            for angle in labels[v1].keys():
                if len(Set(angle).intersection(new_neighbors_set)) == 2:
                    angles_out_of_v1.add(angle)
                    l = l + list(angle)
            # find a unique element in l
            l.sort()
            i = 0
            while i < len(l):
                if l[i] == l[i + 1]:
                    i += 2
                else:
                    break

            angle_set = Set(angles_out_of_v1)

            vertices_in_order.append(l[i])
            while angles_out_of_v1:
                for angle in angles_out_of_v1:
                    if vertices_in_order[-1] in angle:
                        break
                if angle[0] == vertices_in_order[-1]:
                    vertices_in_order.append(angle[1])
                else:
                    vertices_in_order.append(angle[0])
                angles_out_of_v1.remove(angle)

            w = vertices_in_order

            # is w[0] a 2 or a 3?
            top_label = labels[w[0]][tuple(sorted((v1, w[1])))]
            if top_label == 3:
                bottom_label = 2
            else:
                bottom_label = 3
            i = 0
            while i < len(w) - 1:
                labels[v][tuple(sorted((w[i], w[i + 1])))] = 1
                labels[w[i]][tuple(sorted((w[i + 1], v)))] = top_label
                labels[w[i + 1]][tuple(sorted((w[i], v)))] = bottom_label
                i += 1

            labels[v][tuple(sorted((v1, w[0])))] = bottom_label
            labels[v][tuple(sorted((v1, w[-1])))] = top_label

            labels[w[0]][tuple(sorted((v1, v)))] = top_label
            labels[w[-1]][tuple(sorted((v1, v)))] = bottom_label
            labels[v1][tuple(sorted((w[0], v)))] = 1
            labels[v1][tuple(sorted((w[-1], v)))] = 1

            # delete all the extra labels

            for angle in angle_set:
                labels[v1].pop(angle)

            labels[w[0]].pop(tuple(sorted((v1, w[1]))))
            labels[w[-1]].pop(tuple(sorted((v1, w[-2]))))

            i = 1
            while i < len(w) - 1:
                labels[w[i]].pop(tuple(sorted((v1, w[i + 1]))))
                labels[w[i]].pop(tuple(sorted((v1, w[i - 1]))))
                i += 1

        for w in new_neighbors:
            g.add_edge((v, w))

    # Up to this point we did not take the order of the external face into
    # account. Since the combinatorial embedding of a triangulation is unique up
    # to the choice of the outer face and reflection, this might lead to a
    # reflection of the Schnyder drawing resulting from this labeling which is
    # not conformal with comb_emb any longer. Therefore, we might have to swap
    # the labels 1 and 2.
    if (v1, v2) in external_face:
        for u in labels:
            for v, w in labels[u]:
                if labels[u][v, w] == 1:
                    labels[u][v, w] = 2
                elif labels[u][v, w] == 2:
                    labels[u][v, w] = 1
        v1, v2 = v2, v1

    return labels, (v1, v2, v3)


def _realizer(g, x, example=False):
    """
    Given a triangulated graph g and a normal labeling constructs the
    realizer and returns a dictionary of three trees determined by the
    realizer, each spanning all interior vertices and rooted at one of
    the three external vertices.

    A realizer is a directed graph with edge labels that span all interior
    vertices from each external vertex.  It is determined by giving direction
    to the edges that have the same angle label on both sides at a vertex.
    (Thus the direction actually points to the parent in the tree.)  The
    edge label is set as whatever the matching angle label is.  Then from
    any interior vertex, following the directed edges by label will
    give a path to each of the three external vertices.

    INPUT:

    - ``g`` -- the graph to compute the realizer of
    - ``x`` -- tuple with entries

        x[0] = dict of dicts representing a normal labeling of g.  For
        each vertex of g and each adjacent neighbors u,v (u < v) of
        vertex:  { vertex : { (u,v): angle_label } }

        x[1] = (v1, v2, v3) tuple of the three external vertices (also
        the roots of each tree)

    OUTPUT: x; tuple with entries

        x[0] = dict of lists of TreeNodes:

        { root_vertex : [ list of all TreeNodes under root_vertex ] }

        x[1] = (v1,v2,v3) tuple of the three external vertices (also the
        roots of each tree)

    EXAMPLES::

        sage: from sage.graphs.schnyder import _triangulate, _normal_label, _realizer
        sage: g = Graph(graphs.CycleGraph(7))
        sage: g.is_planar(set_embedding=True)
        True
        sage: faces = g.faces(g._embedding)
        sage: _triangulate(g, g._embedding)
        [(2, 0), (4, 2), (6, 4), (5, 0), (3, 5), (1, 3), (4, 0), (3, 0)]
        sage: tn = _normal_label(g, g._embedding, faces[0])
        sage: _realizer(g, tn)
        ({0: [<sage.graphs.schnyder.TreeNode object at ...>]},
         (1, 0, 2))
    """
    normal_labeling, (v1, v2, v3) = x
    realizer = DiGraph()

    tree_nodes = {}
    for v in g:
        tree_nodes[v] = [TreeNode(label=v, children=[]),
                         TreeNode(label=v, children=[]),
                         TreeNode(label=v, children=[])]

    for v in g:
        ones = []
        twos = []
        threes = []
        l = [ones, twos, threes]
        for angle, value in normal_labeling[v].items():
            l[value - 1] += list(angle)

        ones.sort()
        twos.sort()
        threes.sort()

        i = 0
        while i < len(ones) - 1:
            if ones[i] == ones[i + 1]:
                realizer.add_edge((ones[i], v), label=1)
                tree_nodes[v][0].append_child(tree_nodes[ones[i]][0])
                i += 1
            i += 1
        i = 0
        while i < len(twos) - 1:
            if twos[i] == twos[i + 1]:
                realizer.add_edge((twos[i], v), label=2)
                tree_nodes[v][1].append_child(tree_nodes[twos[i]][1])
                i += 1
            i += 1
        i = 0
        while i < len(threes) - 1:
            if threes[i] == threes[i + 1]:
                realizer.add_edge((threes[i], v), label=3)
                tree_nodes[v][2].append_child(tree_nodes[threes[i]][2])
                i += 1
            i += 1

    _compute_coordinates(realizer, (tree_nodes, (v1, v2, v3)))

    if example:
        realizer.show(talk=True, edge_labels=True)

    return tree_nodes, (v1, v2, v3)


def _compute_coordinates(g, x):
    r"""
    Given a triangulated graph g with a dict of trees given by the
    realizer and tuple of the external vertices, we compute the
    coordinates of a planar geometric embedding in the grid.

    The coordinates will be set to the ``_pos`` attribute of g.

    INPUT:

    - ``g`` -- the graph to compute the coordinates of
    - ``x`` -- tuple with entries

        x[0] = dict of tree nodes for the three trees with each external
        vertex as root:

        { root_vertex : [ list of all TreeNodes under root_vertex ] }

        x[1] = (v1, v2, v3) tuple of the three external vertices (also
        the roots of each tree)

    EXAMPLES::

        sage: from sage.graphs.schnyder import _triangulate, _normal_label, _realizer, _compute_coordinates
        sage: g = Graph(graphs.CycleGraph(7))
        sage: g.is_planar(set_embedding=True)
        True
        sage: faces = g.faces(g._embedding)
        sage: _triangulate(g, g._embedding)
        [(2, 0), (4, 2), (6, 4), (5, 0), (3, 5), (1, 3), (4, 0), (3, 0)]
        sage: tn = _normal_label(g, g._embedding, faces[0])
        sage: r = _realizer(g, tn)
        sage: _compute_coordinates(g,r)
        sage: g.get_pos()
        {0: [0, 5], 1: [5, 1], 2: [1, 0], 3: [4, 1], 4: [1, 1], 5: [2, 2], 6: [1, 2]}
    """

    tree_nodes, (v1, v2, v3) = x
    # find the roots of each tree:
    t1, t2, t3 = tree_nodes[v1][0], tree_nodes[v2][1], tree_nodes[v3][2]

    # Compute the number of descendants and depth of each node in
    # each tree.
    t1.compute_number_of_descendants()
    t2.compute_number_of_descendants()
    t3.compute_number_of_descendants()

    t1.compute_depth_of_self_and_children()
    t2.compute_depth_of_self_and_children()
    t3.compute_depth_of_self_and_children()

    coordinates = {}  # the dict to pass to g.set_pos()

    # Setting coordinates for external vertices
    coordinates[t1.label] = [g.order() - 2, 1]
    coordinates[t2.label] = [0, g.order() - 2]
    coordinates[t3.label] = [1, 0]

    for v in g.vertices(sort=False):
        if v not in [t1.label, t2.label, t3.label]:
            # Computing coordinates for v
            r = [0, 0, 0]

            for i in [0, 1, 2]:
                # Computing size of region i:

                # Tracing up tree (i + 1) % 3
                p = tree_nodes[v][(i + 1) % 3]
                while p is not None:
                    q = tree_nodes[p.label][i].number_of_descendants
                    # Adding number of descendants from Tree i nodes with
                    # labels on path up tree (i + 1) % 3
                    r[i] += q
                    p = p.parent

                # Tracing up tree (i - 1) % 3
                p = tree_nodes[v][(i - 1) % 3]
                while p is not None:
                    q = tree_nodes[p.label][i].number_of_descendants
                    # Adding number of descendants from Tree i nodes with
                    # labels on path up tree (i - 1) % 3
                    r[i] += q
                    p = p.parent

                q = tree_nodes[v][i].number_of_descendants
                # Subtracting
                r[i] -= q

                # Subtracting
                q = tree_nodes[v][(i - 1) % 3].depth
                r[i] -= q

            if sum(r) != g.order() - 1:
                raise RuntimeError("Computing coordinates failed: vertex %s's coordinates sum to %s.  Expected %s" % (v, sum(r), g.order() - 1))

            coordinates[v] = r[:-1]

    g.set_pos(coordinates)  # Setting _pos attribute to store coordinates


class TreeNode:
    """
    A class to represent each node in the trees used by ``_realizer`` and
    ``_compute_coordinates`` when finding a planar geometric embedding in
    the grid.

    Each tree node is doubly linked to its parent and children.

    INPUT:

    - ``parent`` -- the parent TreeNode of ``self``
    - ``children`` -- list of TreeNode children of ``self``
    - ``label`` -- the associated realizer vertex label

    EXAMPLES::

        sage: from sage.graphs.schnyder import TreeNode
        sage: tn = TreeNode(label=5)
        sage: tn2 = TreeNode(label=2,parent=tn)
        sage: tn3 = TreeNode(label=3)
        sage: tn.append_child(tn3)
        sage: tn.compute_number_of_descendants()
        2
        sage: tn.number_of_descendants
        2
        sage: tn3.number_of_descendants
        1
        sage: tn.compute_depth_of_self_and_children()
        sage: tn3.depth
        2
    """
    def __init__(self, parent=None, children=None, label=None):
        """
        INPUT:

        - ``parent`` -- the parent TreeNode of ``self``
        - ``children`` -- list of TreeNode children of ``self``
        - ``label`` -- the associated realizer vertex label

        EXAMPLES::

            sage: from sage.graphs.schnyder import TreeNode
            sage: tn = TreeNode(label=5)
            sage: tn2 = TreeNode(label=2,parent=tn)
            sage: tn3 = TreeNode(label=3)
            sage: tn.append_child(tn3)
            sage: tn.compute_number_of_descendants()
            2
            sage: tn.number_of_descendants
            2
            sage: tn3.number_of_descendants
            1
            sage: tn.compute_depth_of_self_and_children()
            sage: tn3.depth
            2
        """
        if children is None:
            children = []
        self.parent = parent
        self.children = children
        self.label = label
        self.number_of_descendants = 1

    def compute_number_of_descendants(self):
        """
        Compute the number of descendants of ``self`` and all descendants.

        For each TreeNode, sets result as attribute ``self.number_of_descendants``.

        EXAMPLES::

            sage: from sage.graphs.schnyder import TreeNode
            sage: tn = TreeNode(label=5)
            sage: tn2 = TreeNode(label=2,parent=tn)
            sage: tn3 = TreeNode(label=3)
            sage: tn.append_child(tn3)
            sage: tn.compute_number_of_descendants()
            2
            sage: tn.number_of_descendants
            2
            sage: tn3.number_of_descendants
            1
            sage: tn.compute_depth_of_self_and_children()
            sage: tn3.depth
            2
        """
        n = 1
        for child in self.children:
            n += child.compute_number_of_descendants()
        self.number_of_descendants = n
        return n

    def compute_depth_of_self_and_children(self):
        """
        Compute the depth of ``self`` and all descendants.

        For each TreeNode, sets result as ``attribute self.depth``.

        EXAMPLES::

            sage: from sage.graphs.schnyder import TreeNode
            sage: tn = TreeNode(label=5)
            sage: tn2 = TreeNode(label=2,parent=tn)
            sage: tn3 = TreeNode(label=3)
            sage: tn.append_child(tn3)
            sage: tn.compute_number_of_descendants()
            2
            sage: tn.number_of_descendants
            2
            sage: tn3.number_of_descendants
            1
            sage: tn.compute_depth_of_self_and_children()
            sage: tn3.depth
            2
        """
        if self.parent is None:
            self.depth = 1
        else:
            self.depth = self.parent.depth + 1
        for child in self.children:
            child.compute_depth_of_self_and_children()

    def append_child(self, child):
        """
        Add a child to list of children.

        EXAMPLES::

            sage: from sage.graphs.schnyder import TreeNode
            sage: tn = TreeNode(label=5)
            sage: tn2 = TreeNode(label=2,parent=tn)
            sage: tn3 = TreeNode(label=3)
            sage: tn.append_child(tn3)
            sage: tn.compute_number_of_descendants()
            2
            sage: tn.number_of_descendants
            2
            sage: tn3.number_of_descendants
            1
            sage: tn.compute_depth_of_self_and_children()
            sage: tn3.depth
            2
        """
        if child in self.children:
            return
        self.children.append(child)
        child.parent = self


def minimal_schnyder_wood(graph, root_edge=None, minimal=True, check=True):
    """
    Return the minimal Schnyder wood of a planar rooted triangulation.

    INPUT:

    - ``graph`` -- a planar triangulation, given by a graph with an embedding

    - ``root_edge`` -- a pair of vertices (default: from ``-1`` to ``-2``);
      the third boundary vertex is then determined using the orientation and
      will be labelled ``-3``

    - ``minimal`` -- boolean (default: ``True``); whether to return a
      minimal or a maximal Schnyder wood

    - ``check`` -- boolean (default: ``True``); whether to check if the input
      is a planar triangulation

    OUTPUT:

    A planar graph, with edges oriented and colored. The three outer
    edges of the initial graph are removed. For the three outer vertices the
    list of the neighbors stored in the combinatorial embedding is in the order
    of the incident edges between the two incident (and removed) outer edges,
    and not a cyclic shift of it.

    The algorithm is taken from [Bre2000]_ (section 4.2).

    EXAMPLES::

        sage: from sage.graphs.schnyder import minimal_schnyder_wood
        sage: g = Graph([(0,-1),(0,-2),(0,-3),(-1,-2),(-2,-3),
        ....:  (-3,-1)], format='list_of_edges')
        sage: g.set_embedding({-1:[-2,0,-3],-2:[-3,0,-1],
        ....:  -3:[-1,0,-2],0:[-1,-2,-3]})
        sage: newg = minimal_schnyder_wood(g)
        sage: newg.edges(sort=True)
        [(0, -3, 'red'), (0, -2, 'blue'), (0, -1, 'green')]
        sage: newg.plot(color_by_label={'red':'red','blue':'blue',                      # needs sage.plot
        ....:  'green':'green',None:'black'})
        Graphics object consisting of 8 graphics primitives

    A larger example::

        sage: g = Graph([(0,-1),(0,2),(0,1),(0,-3),(-1,-3),(-1,2),
        ....: (-1,-2),(1,2),(1,-3),(2,-2),(1,-2),(-2,-3)], format='list_of_edges')
        sage: g.set_embedding({-1:[-2,2,0,-3],-2:[-3,1,2,-1],
        ....: -3:[-1,0,1,-2],0:[-1,2,1,-3],1:[-2,-3,0,2],2:[-1,-2,1,0]})
        sage: newg = minimal_schnyder_wood(g)
        sage: newg.edges(sort=True, key=lambda e:(str(e[0]),str(e[1])))
        [(0, -1, 'green'),
         (0, -3, 'red'),
         (0, 2, 'blue'),
         (1, -2, 'blue'),
         (1, -3, 'red'),
         (1, 0, 'green'),
         (2, -1, 'green'),
         (2, -2, 'blue'),
         (2, 1, 'red')]
        sage: newg2 = minimal_schnyder_wood(g, minimal=False)
        sage: newg2.edges(sort=True, key=lambda e:(str(e[0]),str(e[1])))
        [(0, -1, 'green'),
         (0, -3, 'red'),
         (0, 1, 'blue'),
         (1, -2, 'blue'),
         (1, -3, 'red'),
         (1, 2, 'green'),
         (2, -1, 'green'),
         (2, -2, 'blue'),
         (2, 0, 'red')]

    TESTS::

        sage: minimal_schnyder_wood(graphs.RandomTriangulation(5))
        Digraph on 5 vertices
        sage: minimal_schnyder_wood(graphs.CompleteGraph(5))
        Traceback (most recent call last):
        ...
        ValueError: not a planar graph
        sage: minimal_schnyder_wood(graphs.WheelGraph(5))
        Traceback (most recent call last):
        ...
        ValueError: not a triangulation
        sage: minimal_schnyder_wood(graphs.OctahedralGraph(),root_edge=(0,5))
        Traceback (most recent call last):
        ...
        ValueError: not a valid root edge
    """
    if root_edge is None:
        a = -1
        b = -2
    else:
        a, b = root_edge

    if check:
        if not graph.is_planar():
            raise ValueError('not a planar graph')
        if not all(len(u) == 3 for u in graph.faces()):
            raise ValueError('not a triangulation')
        if a not in graph.neighbors(b):
            raise ValueError('not a valid root edge')

    new_g = DiGraph()
    emb = graph.get_embedding()

    # finding the third outer vertex c
    emb_b = emb[b]
    idx_a = emb_b.index(a)
    c = emb_b[(idx_a + 1) % len(emb_b)]

    # initialisation
    for i in emb[c]:
        if i != a and i != b:
            new_g.add_edge((i, -3, 'red'))

    path = list(emb[c])
    idxa = path.index(a)
    path = path[idxa:] + path[:idxa]
    neighbors_in_path = {i: len([u for u in graph.neighbors(i) if u in path])
                         for i in graph}
    removable_nodes = [u for u in path if neighbors_in_path[u] == 2 and
                       u != a and u != b]

    # iterated path shortening
    while len(path) > 2:
        if minimal:
            v = removable_nodes[-1]   # node to be removed from path
        else:
            v = removable_nodes[0]   # node to be removed from path
        idx_v = path.index(v)
        left = path[idx_v - 1]
        new_g.add_edge((v, left, 'green'))
        right = path[idx_v + 1]
        new_g.add_edge((v, right, 'blue'))
        neighbors_v = emb[v]
        idx_left = neighbors_v.index(left)
        neighbors_v = neighbors_v[idx_left:] + neighbors_v[:idx_left]
        idx_right = neighbors_v.index(right)
        inside = neighbors_v[1:idx_right]
        new_g.add_edges([(w, v, 'red') for w in inside])
        path = path[:idx_v] + inside + path[idx_v + 1:]
        # updating the table of neighbors_in_path
        for w in inside:
            for x in graph.neighbors(w):
                neighbors_in_path[x] += 1
        for x in graph.neighbors(v):
            neighbors_in_path[x] -= 1
        # updating removable nodes
        removable_nodes = [u for u in path if neighbors_in_path[u] == 2 and
                           u != a and u != b]

    def relabel(w):
        return -3 if w == c else w

    emb = {relabel(v): [relabel(u) for u in emb[v]] for v in graph}
    for u, v, w in (a, b, -3), (b, -3, a), (-3, a, b):
        idx = emb[u].index(v)
        if idx == 0:
            emb[u] = emb[u][1:-1]
        else:
            emb[u] = emb[u][idx+1:] + emb[u][:idx-1]

    new_g.set_embedding(emb)
    return new_g



# sage_setup: distribution = sagemath-graphs
"""
Cartesian products of posets

AUTHORS:

- Daniel Krenn (2015)
"""
# ****************************************************************************
#  Copyright (C) 2015 Daniel Krenn <dev@danielkrenn.at>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                https://www.gnu.org/licenses/
# ****************************************************************************

from sage.sets.cartesian_product import CartesianProduct


class CartesianProductPoset(CartesianProduct):
    r"""
    A class implementing Cartesian products of posets (and elements
    thereof). Compared to :class:`CartesianProduct` you are able to
    specify an order for comparison of the elements.

    INPUT:

    - ``sets`` -- tuple of parents

    - ``category`` -- a subcategory of ``Sets().CartesianProducts() & Posets()``

    - ``order`` -- string or function specifying an order less or equal;
      it can be one of the following:

      - ``'native'`` -- elements are ordered by their native ordering,
        i.e., the order the wrapped elements (tuples) provide

      - ``'lex'`` -- elements are ordered lexicographically

      - ``'product'`` -- an element is less or equal to another
        element, if less or equal is true for all its components
        (Cartesian projections)

      - a function which performs the comparison `\leq`; it takes two
        input arguments and outputs a boolean

    Other keyword arguments (``kwargs``) are passed to the constructor
    of :class:`CartesianProduct`.

    EXAMPLES::

        sage: P = Poset((srange(3), lambda left, right: left <= right))
        sage: Cl = cartesian_product((P, P), order='lex')
        sage: Cl((1, 1)) <= Cl((2, 0))
        True
        sage: Cp = cartesian_product((P, P), order='product')
        sage: Cp((1, 1)) <= Cp((2, 0))
        False
        sage: def le_sum(left, right):
        ....:     return (sum(left) < sum(right) or
        ....:             sum(left) == sum(right) and left[0] <= right[0])
        sage: Cs = cartesian_product((P, P), order=le_sum)
        sage: Cs((1, 1)) <= Cs((2, 0))
        True

    TESTS::

        sage: Cl.category()
        Join of Category of finite posets and
        Category of Cartesian products of finite enumerated sets
        sage: TestSuite(Cl).run(skip=['_test_construction'])
        sage: Cp.category()
        Join of Category of finite posets and
        Category of Cartesian products of finite enumerated sets
        sage: TestSuite(Cp).run(skip=['_test_construction'])

    .. SEEALSO::

        :class:`CartesianProduct`
    """

    def __init__(self, sets, category, order=None, **kwargs) -> None:
        r"""
        See :class:`CartesianProductPoset` for details.

        TESTS::

            sage: P = Poset((srange(3), lambda left, right: left <= right))
            sage: C = cartesian_product((P, P), order='notexisting')
            Traceback (most recent call last):
            ...
            ValueError: no order 'notexisting' known
            sage: C = cartesian_product((P, P), category=(Groups(),))
            sage: C.category()
            Join of Category of groups and Category of posets
        """
        if order is None:
            self._le_ = self.le_product
        elif isinstance(order, str):
            try:
                self._le_ = getattr(self, 'le_' + order)
            except AttributeError:
                raise ValueError(f"no order '{order}' known")
        else:
            self._le_ = order

        from sage.categories.category import Category
        from sage.categories.posets import Posets
        if not isinstance(category, tuple):
            category = (category,)
        category = Category.join(category + (Posets(),))
        super().__init__(sets, category, **kwargs)

    def le(self, left, right):
        r"""
        Test whether ``left`` is less than or equal to ``right``.

        INPUT:

        - ``left`` -- an element

        - ``right`` -- an element

        OUTPUT: boolean

        .. NOTE::

            This method uses the order defined on creation of this
            Cartesian product. See :class:`CartesianProductPoset`.

        EXAMPLES::

            sage: P = posets.ChainPoset(10)
            sage: def le_sum(left, right):
            ....:     return (sum(left) < sum(right) or
            ....:             sum(left) == sum(right) and left[0] <= right[0])
            sage: C = cartesian_product((P, P), order=le_sum)
            sage: C.le(C((1, 6)), C((6, 1)))
            True
            sage: C.le(C((6, 1)), C((1, 6)))
            False
            sage: C.le(C((1, 6)), C((6, 6)))
            True
            sage: C.le(C((6, 6)), C((1, 6)))
            False
        """
        return self._le_(left, right)

    def le_lex(self, left, right):
        r"""
        Test whether ``left`` is lexicographically smaller or equal
        to ``right``.

        INPUT:

        - ``left`` -- an element

        - ``right`` -- an element

        OUTPUT: boolean

        EXAMPLES::

            sage: P = Poset((srange(2), lambda left, right: left <= right))
            sage: Q = cartesian_product((P, P), order='lex')
            sage: T = [Q((0, 0)), Q((1, 1)), Q((0, 1)), Q((1, 0))]
            sage: for a in T:
            ....:     for b in T:
            ....:         assert Q.le(a, b) == (a <= b)
            ....:         print('%s <= %s = %s' % (a, b, a <= b))
            (0, 0) <= (0, 0) = True
            (0, 0) <= (1, 1) = True
            (0, 0) <= (0, 1) = True
            (0, 0) <= (1, 0) = True
            (1, 1) <= (0, 0) = False
            (1, 1) <= (1, 1) = True
            (1, 1) <= (0, 1) = False
            (1, 1) <= (1, 0) = False
            (0, 1) <= (0, 0) = False
            (0, 1) <= (1, 1) = True
            (0, 1) <= (0, 1) = True
            (0, 1) <= (1, 0) = True
            (1, 0) <= (0, 0) = False
            (1, 0) <= (1, 1) = True
            (1, 0) <= (0, 1) = False
            (1, 0) <= (1, 0) = True

        TESTS:

        Check that :issue:`19999` is resolved::

            sage: P = Poset((srange(2), lambda left, right: left <= right))
            sage: Q = cartesian_product((P, P), order='product')
            sage: R = cartesian_product((Q, P), order='lex')
            sage: R(((1, 0), 0)) <= R(((0, 1), 0))
            False
            sage: R(((0, 1), 0)) <= R(((1, 0), 0))
            False
        """
        for l, r, S in \
                zip(left.value, right.value, self.cartesian_factors()):
            if l == r:
                continue
            if S.le(l, r):
                return True
            if S.le(r, l):
                return False
            return False  # incomparable components
        return True  # equal

    def le_product(self, left, right):
        r"""
        Test whether ``left`` is component-wise smaller or equal
        to ``right``.

        INPUT:

        - ``left`` -- an element

        - ``right`` -- an element

        OUTPUT: boolean

        The comparison is ``True`` if the result of the
        comparison in each component is ``True``.

        EXAMPLES::

            sage: P = Poset((srange(2), lambda left, right: left <= right))
            sage: Q = cartesian_product((P, P), order='product')
            sage: T = [Q((0, 0)), Q((1, 1)), Q((0, 1)), Q((1, 0))]
            sage: for a in T:
            ....:     for b in T:
            ....:         assert Q.le(a, b) == (a <= b)
            ....:         print('%s <= %s = %s' % (a, b, a <= b))
            (0, 0) <= (0, 0) = True
            (0, 0) <= (1, 1) = True
            (0, 0) <= (0, 1) = True
            (0, 0) <= (1, 0) = True
            (1, 1) <= (0, 0) = False
            (1, 1) <= (1, 1) = True
            (1, 1) <= (0, 1) = False
            (1, 1) <= (1, 0) = False
            (0, 1) <= (0, 0) = False
            (0, 1) <= (1, 1) = True
            (0, 1) <= (0, 1) = True
            (0, 1) <= (1, 0) = False
            (1, 0) <= (0, 0) = False
            (1, 0) <= (1, 1) = True
            (1, 0) <= (0, 1) = False
            (1, 0) <= (1, 0) = True
        """
        return all(
            S.le(l, r)
            for l, r, S in
            zip(left.value, right.value, self.cartesian_factors()))

    def le_native(self, left, right):
        r"""
        Test whether ``left`` is smaller or equal to ``right`` in the order
        provided by the elements themselves.

        INPUT:

        - ``left`` -- an element

        - ``right`` -- an element

        OUTPUT: boolean

        EXAMPLES::

            sage: P = Poset((srange(2), lambda left, right: left <= right))
            sage: Q = cartesian_product((P, P), order='native')
            sage: T = [Q((0, 0)), Q((1, 1)), Q((0, 1)), Q((1, 0))]
            sage: for a in T:
            ....:     for b in T:
            ....:         assert Q.le(a, b) == (a <= b)
            ....:         print('%s <= %s = %s' % (a, b, a <= b))
            (0, 0) <= (0, 0) = True
            (0, 0) <= (1, 1) = True
            (0, 0) <= (0, 1) = True
            (0, 0) <= (1, 0) = True
            (1, 1) <= (0, 0) = False
            (1, 1) <= (1, 1) = True
            (1, 1) <= (0, 1) = False
            (1, 1) <= (1, 0) = False
            (0, 1) <= (0, 0) = False
            (0, 1) <= (1, 1) = True
            (0, 1) <= (0, 1) = True
            (0, 1) <= (1, 0) = True
            (1, 0) <= (0, 0) = False
            (1, 0) <= (1, 1) = True
            (1, 0) <= (0, 1) = False
            (1, 0) <= (1, 0) = True
        """
        return left.value <= right.value

    class Element(CartesianProduct.Element):

        def _le_(self, other):
            r"""
            Return if this element is less or equal to ``other``.

            INPUT:

            - ``other`` -- an element

            OUTPUT: boolean

            .. NOTE::

                This method calls :meth:`CartesianProductPoset.le`. Override
                it in inherited class to change this.

                It can be assumed that this element and ``other`` have
                the same parent.

            TESTS::

                sage: from sage.combinat.posets.cartesian_product import CartesianProductPoset
                sage: QQ.CartesianProduct = CartesianProductPoset  # needed until #19269 is fixed
                sage: def le_sum(left, right):
                ....:     return (sum(left) < sum(right) or
                ....:             sum(left) == sum(right) and left[0] <= right[0])
                sage: C = cartesian_product((QQ, QQ), order=le_sum)
                sage: C((1/3, 2)) <= C((2, 1/3))  # indirect doctest
                True
                sage: C((1/3, 2)) <= C((2, 2))  # indirect doctest
                True
            """
            return self.parent().le(self, other)

        def __le__(self, other):
            r"""
            Return if this element is less than or equal to ``other``.

            INPUT:

            - ``other`` -- an element

            OUTPUT: boolean

            .. NOTE::

                This method uses the coercion framework to find a
                suitable common parent.

                This method can be deleted once :issue:`10130` is fixed and
                provides these methods automatically.

            TESTS::

                sage: from sage.combinat.posets.cartesian_product import CartesianProductPoset
                sage: QQ.CartesianProduct = CartesianProductPoset  # needed until #19269 is fixed
                sage: def le_sum(left, right):
                ....:     return (sum(left) < sum(right) or
                ....:             sum(left) == sum(right) and left[0] <= right[0])
                sage: C = cartesian_product((QQ, QQ), order=le_sum)
                sage: C((1/3, 2)) <= C((2, 1/3))
                True
                sage: C((1/3, 2)) <= C((2, 2))
                True

            The following example tests that the coercion gets involved in
            comparisons; it can be simplified once :issue:`18182` is merged.
            ::

                sage: class MyCP(CartesianProductPoset):
                ....:     def _coerce_map_from_(self, S):
                ....:         if isinstance(S, self.__class__):
                ....:             S_factors = S.cartesian_factors()
                ....:             R_factors = self.cartesian_factors()
                ....:             if len(S_factors) == len(R_factors):
                ....:                 if all(r.has_coerce_map_from(s)
                ....:                        for r,s in zip(R_factors, S_factors)):
                ....:                     return True
                sage: QQ.CartesianProduct = MyCP
                sage: A = cartesian_product((QQ, ZZ), order=le_sum)
                sage: B = cartesian_product((QQ, QQ), order=le_sum)
                sage: A((1/2, 4)) <= B((1/2, 5))
                True
            """
            from sage.structure.element import have_same_parent
            if have_same_parent(self, other):
                return self._le_(other)

            from sage.structure.element import get_coercion_model
            import operator
            try:
                return get_coercion_model().bin_op(self, other, operator.le)
            except TypeError:
                return False

        def __ge__(self, other):
            r"""
            Return if this element is greater than or equal to ``other``.

            INPUT:

            - ``other`` -- an element

            OUTPUT: boolean

            .. NOTE::

                This method uses the coercion framework to find a
                suitable common parent.

                This method can be deleted once :issue:`10130` is fixed and
                provides these methods automatically.

            TESTS::

                sage: from sage.combinat.posets.cartesian_product import CartesianProductPoset
                sage: QQ.CartesianProduct = CartesianProductPoset  # needed until #19269 is fixed
                sage: def le_sum(left, right):
                ....:     return (sum(left) < sum(right) or
                ....:             sum(left) == sum(right) and left[0] <= right[0])
                sage: C = cartesian_product((QQ, QQ), order=le_sum)
                sage: C((1/3, 2)) >= C((2, 1/3))
                False
                sage: C((1/3, 2)) >= C((2, 2))
                False
            """
            return other <= self

        def __lt__(self, other):
            r"""
            Return if this element is less than ``other``.

            INPUT:

            - ``other`` -- an element

            OUTPUT: boolean

            .. NOTE::

                This method uses the coercion framework to find a
                suitable common parent.

                This method can be deleted once :issue:`10130` is fixed and
                provides these methods automatically.

            TESTS::

                sage: from sage.combinat.posets.cartesian_product import CartesianProductPoset
                sage: QQ.CartesianProduct = CartesianProductPoset  # needed until #19269 is fixed
                sage: def le_sum(left, right):
                ....:     return (sum(left) < sum(right) or
                ....:             sum(left) == sum(right) and left[0] <= right[0])
                sage: C = cartesian_product((QQ, QQ), order=le_sum)
                sage: C((1/3, 2)) < C((2, 1/3))
                True
                sage: C((1/3, 2)) < C((2, 2))
                True
            """
            return not self == other and self <= other

        def __gt__(self, other):
            r"""
            Return if this element is greater than ``other``.

            INPUT:

            - ``other`` -- an element

            OUTPUT: boolean

            .. NOTE::

                This method uses the coercion framework to find a
                suitable common parent.

                This method can be deleted once :issue:`10130` is fixed and
                provides these methods automatically.

            TESTS::

                sage: from sage.combinat.posets.cartesian_product import CartesianProductPoset
                sage: QQ.CartesianProduct = CartesianProductPoset  # needed until #19269 is fixed
                sage: def le_sum(left, right):
                ....:     return (sum(left) < sum(right) or
                ....:             sum(left) == sum(right) and left[0] <= right[0])
                sage: C = cartesian_product((QQ, QQ), order=le_sum)
                sage: C((1/3, 2)) > C((2, 1/3))
                False
                sage: C((1/3, 2)) > C((2, 2))
                False
            """
            return not self == other and other <= self



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs database_graphs
r"""
ISGCI: Information System on Graph Classes and their Inclusions

This module implements an interface to the
`ISGCI <http://www.graphclasses.org/>`_ database in Sage.

This database gathers information on graph classes and their inclusions in each
other. It also contains information on the complexity of several computational
problems.

It is available on the `GraphClasses.org <http://www.graphclasses.org/>`_
website maintained by H.N. de Ridder et al.

How to use it?
--------------

Presently, it is possible to use this database through the variables and methods
present in the :obj:`graph_classes <GraphClasses>` object.
For instance::

    sage: Trees = graph_classes.Tree
    sage: Chordal = graph_classes.Chordal

Inclusions
^^^^^^^^^^

It is then possible to check the inclusion of classes inside of others, if the
information is available in the database::

    sage: Trees <= Chordal
    True

And indeed, trees are chordal graphs.

The ISGCI database is not all-knowing, and so comparing two classes can return
``True``, ``False``, or ``Unknown`` (see the :mod:`documentation of the Unknown
truth value <sage.misc.unknown>`).

An *unknown* answer to ``A <= B`` only means that ISGCI cannot deduce from the
information in its database that ``A`` is a subclass of ``B`` nor that it is
not. For instance, ISGCI does not know at the moment that some chordal graphs
are not trees::

    sage: graph_classes.Chordal <= graph_classes.Tree
    Unknown

Descriptions
^^^^^^^^^^^^

Given a graph class, one can obtain its associated information in the ISGCI
database with the :meth:`~GraphClass.description` method::

    sage: Chordal.description()
    Class of graphs : Chordal
    -------------------------
    id                             :  gc_32
    name                           :  chordal
    ...
    Problems :
    -----------
    3-Colourability                :  Linear
    Clique                         :  Polynomial
    Clique cover                   :  Polynomial
    ...

It is possible to obtain the complete list of the classes stored in ISGCI by
calling the :meth:`~GraphClasses.show_all` method (beware -- long output)::

    sage: graph_classes.show_all()
    id        | name                                     | type                 | smallgraph
    ----------------------------------------------------------------------------------------------------------------------
    gc_309    | $K_4$--minor--free                       | base                 |
    gc_541    | $N^*$                                    | base                 |
    gc_215    | $N^*$--perfect                           | base                 |
    gc_5      | $P_4$--bipartite                         | base                 |
    gc_3      | $P_4$--brittle                           | base                 |
    gc_6      | $P_4$--comparability                     | base                 |
    gc_7      | $P_4$--extendible                        | base                 |
    ...

Until a proper search method is implemented, this lets one find classes which do
not appear in :obj:`graph_classes.* <GraphClasses>`.

To retrieve a class of graph from its ISGCI ID one may use
the :meth:`~GraphClasses.get_class` method::

    sage: GC = graph_classes.get_class("gc_5")
    sage: GC
    $P_4$--bipartite graphs

Recognition of graphs
^^^^^^^^^^^^^^^^^^^^^

The graph classes represented by the ISGCI database can alternatively be used to
access recognition algorithms. For instance, in order to check that a given
graph is a tree one has the following the options ::

    sage: graphs.PathGraph(5) in graph_classes.Tree
    True

or::

    sage: graphs.PathGraph(5).is_tree()
    True

Furthermore, all ISGCI graph classes which are defined by the exclusion of a
finite sequence of induced subgraphs benefit from a generic recognition
algorithm. For instance ::

    sage: g = graphs.PetersenGraph()
    sage: g in graph_classes.ClawFree
    False
    sage: g.line_graph() in graph_classes.ClawFree
    True

Or directly from ISGCI ::

    sage: gc = graph_classes.get_class("gc_441")
    sage: gc
    diamond--free graphs
    sage: graphs.PetersenGraph() in gc
    True

Predefined classes
------------------

:obj:`graph_classes <GraphClasses>` currently predefines the following graph classes

.. list-table::
   :widths: 20 30
   :header-rows: 1

   * - Class
     - Related methods

   * - Apex

     - :meth:`~sage.graphs.graph.Graph.is_apex`,
       :meth:`~sage.graphs.graph.Graph.apex_vertices`

   * - AT_free

     - :meth:`~sage.graphs.graph.Graph.is_asteroidal_triple_free`

   * - Biconnected

     - :meth:`~sage.graphs.graph.Graph.is_biconnected`,
       :meth:`~sage.graphs.generic_graph.GenericGraph.blocks_and_cut_vertices`,
       :meth:`~sage.graphs.generic_graph.GenericGraph.blocks_and_cuts_tree`

   * - BinaryTrees

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.BalancedTree`,
       :meth:`~sage.graphs.graph.Graph.is_tree`

   * - Bipartite

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.BalancedTree`,
       :meth:`~sage.graphs.generic_graph.GenericGraph.is_bipartite`

   * - Block

     - :meth:`~sage.graphs.graph.Graph.is_block_graph`,
       :meth:`~sage.graphs.generic_graph.GenericGraph.blocks_and_cut_vertices`,
       :meth:`~sage.graphs.graph_generators.GraphGenerators.RandomBlockGraph`

   * - Chordal

     - :meth:`~sage.graphs.generic_graph.GenericGraph.is_chordal`

   * - Claw-Free
     - :meth:`~sage.graphs.graph_generators.GraphGenerators.ClawGraph`

   * - Comparability
     -

   * - Gallai

     - :meth:`~sage.graphs.generic_graph.GenericGraph.is_gallai_tree`

   * - Grid

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.Grid2dGraph`,
       :meth:`~sage.graphs.graph_generators.GraphGenerators.GridGraph`

   * - Interval

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.RandomIntervalGraph`,
       :meth:`~sage.graphs.graph_generators.GraphGenerators.IntervalGraph`,
       :meth:`~sage.graphs.generic_graph.GenericGraph.is_interval`

   * - Line

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.line_graph_forbidden_subgraphs`,
       :meth:`~sage.graphs.graph.Graph.is_line_graph`

   * - Modular

     - :meth:`~sage.graphs.graph.Graph.modular_decomposition`

   * - Outerplanar

     - :meth:`~sage.graphs.generic_graph.GenericGraph.is_circular_planar`

   * - Perfect

     - :meth:`~sage.graphs.graph.Graph.is_perfect`

   * - Planar

     - :meth:`~sage.graphs.generic_graph.GenericGraph.is_planar`

   * - Polyhedral

     - :meth:`~sage.graphs.graph.Graph.is_polyhedral`

   * - Split

     - :meth:`~sage.graphs.graph.Graph.is_split`

   * - Tree

     - :meth:`~sage.graphs.graph_generators.GraphGenerators.trees`,
       :meth:`~Graph.is_tree`

   * - UnitDisk
     - :meth:`~sage.graphs.graph_generators.GraphGenerators.IntervalGraph`

   * - UnitInterval
     - :meth:`~sage.graphs.generic_graph.GenericGraph.is_interval`

Sage's view of ISGCI
--------------------

The database is stored by Sage in two ways.

**The classes**: the list of all graph classes and their properties is stored
in a huge dictionary (see :meth:`~sage.graphs.isgci.GraphClasses.classes`).
Below is what Sage knows of ``gc_249``::

    sage: graph_classes.classes()['gc_249']        # random
    {'problem':
        {'Independent set': 'Polynomial',
         'Treewidth': 'Unknown',
         'Weighted independent set': 'Polynomial',
         'Cliquewidth expression': 'NP-complete',
         'Weighted clique': 'Polynomial',
         'Clique cover': 'Unknown',
         'Domination': 'NP-complete',
         'Clique': 'Polynomial',
         'Colourability': 'NP-complete',
         'Cliquewidth': 'Unbounded',
         '3-Colourability': 'NP-complete',
         'Recognition': 'Linear'},
     'type': 'base',
     'id': 'gc_249',
     'name': 'line'}

**The class inclusion digraph**: Sage remembers the class inclusions through
the inclusion digraph (see :meth:`~sage.graphs.isgci.GraphClasses.inclusion_digraph`).
Its nodes are ID of ISGCI classes::

    sage: d = graph_classes.inclusion_digraph()
    sage: d.vertices(sort=True)[-10:]
    ['gc_990', 'gc_991', 'gc_992', 'gc_993', 'gc_994', 'gc_995', 'gc_996', 'gc_997', 'gc_998', 'gc_999']

An arc from ``gc1`` to ``gc2`` means that ``gc1`` is a superclass of ``gc2``.
This being said, not all edges are stored ! To ensure that a given class is
included in another one, we have to check whether there is in the digraph a
``path`` from the first one to the other::

    sage: bip_id = graph_classes.Bipartite._gc_id
    sage: perfect_id = graph_classes.Perfect._gc_id
    sage: d.has_edge(perfect_id, bip_id)
    False
    sage: d.distance(perfect_id, bip_id)
    2

Hence bipartite graphs are perfect graphs. We can see how ISGCI obtains this
result ::

    sage: p = d.shortest_path(perfect_id, bip_id)
    sage: len(p) - 1
    2
    sage: print(p)                  # random
    ['gc_56', 'gc_76', 'gc_69']
    sage: for c in p:
    ....:     print(graph_classes.get_class(c))
    perfect graphs
    ...
    bipartite graphs

What ISGCI knows is that perfect graphs contain unimodular graph which contain
bipartite graphs. Therefore bipartite graphs are perfect !

.. NOTE::

    The inclusion digraph is **NOT ACYCLIC**. Indeed, several entries exist in
    the ISGCI database which represent the same graph class, for instance
    Perfect graphs and Berge graphs::

        sage: graph_classes.inclusion_digraph().is_directed_acyclic()
        False
        sage: Berge = graph_classes.get_class("gc_274"); Berge
        Berge graphs
        sage: Perfect = graph_classes.get_class("gc_56"); Perfect
        perfect graphs
        sage: Berge <= Perfect
        True
        sage: Perfect <= Berge
        True
        sage: Perfect == Berge
        True

Information for developers
--------------------------

* The database is loaded not *so* large, but it is still preferable to
  only load it on demand. This is achieved through the cached methods
  :meth:`~sage.graphs.isgci.GraphClasses.classes` and
  :meth:`~sage.graphs.isgci.GraphClasses.inclusion_digraph`.

* Upon the first access to the database, the information is extracted
  from the XML file and stored in the cache of three methods:

  * ``sage.graphs.isgci._classes`` (dictionary)
  * ``sage.graphs.isgci._inclusions`` (list of dictionaries)
  * ``sage.graphs.isgci._inclusion_digraph`` (DiGraph)

  Note that the digraph is only built if necessary (for instance if
  the user tries to compare two classes).

.. TODO::

    Technical things:

    * Query the database for non-inclusion results so that comparisons can
      return ``False``, and implement strict inclusions.

    * Implement a proper search method for the classes not listed in
      :obj:`graph_classes <GraphClasses>`

      .. SEEALSO:: :func:`sage.graphs.isgci.show_all`.

    * Some of the graph classes appearing in :obj:`graph_classes
      <GraphClasses>` already have a recognition
      algorithm implemented in Sage. It would be so nice to be able to
      write ``g in Trees``, ``g in Perfect``, ``g in Chordal``, ... :-)

    Long-term stuff:

    * Implement simple accessors for all the information in the ISGCI
      database (as can be done from the website)

    * Implement intersection of graph classes

    * Write generic recognition algorithms for specific classes (when a graph
      class is defined by the exclusion of subgraphs, one can write a generic
      algorithm checking the existence of each of the graphs, and this method
      already exists in Sage).

    * Improve the performance of Sage's graph library by letting it take
      advantage of the properties of graph classes. For example,
      :meth:`Graph.independent_set` could use the library to detect that a given
      graph is, say, a tree or a planar graph, and use a specialized algorithm
      for finding an independent set.

AUTHORS:
--------

* H.N. de Ridder et al. (ISGCI database)
* Nathann Cohen (Sage implementation)

Methods
-------
"""

from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import CachedRepresentation, UniqueRepresentation
from sage.misc.unknown import Unknown
from sage.features.databases import DatabaseGraphs
from sage.misc.cachefunc import cached_method

import os
import zipfile
from urllib.request import urlopen
from ssl import create_default_context as default_context

# ****************************************************************************
#      Copyright (C) 2011 Nathann Cohen <nathann.cohen@gmail.com>
#
# Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
#                         https://www.gnu.org/licenses/
# ****************************************************************************

_XML_FILE = "isgci_sage.xml"
_SMALLGRAPHS_FILE = "smallgraphs.txt"


class GraphClass(SageObject, CachedRepresentation):
    r"""
    An instance of this class represents a Graph Class, matching some entry in
    the ISGCI database.

    EXAMPLES:

    Testing the inclusion of two classes::

        sage: Chordal = graph_classes.Chordal
        sage: Trees = graph_classes.Tree
        sage: Trees <= Chordal
        True
        sage: Chordal <= Trees
        Unknown

    TESTS::

        sage: Trees >= Chordal
        Unknown
        sage: Chordal >= Trees
        True
    """
    def __init__(self, name, gc_id, recognition_function=None):
        r"""
        Class constructor.

        INPUT:

        - ``gc_id`` -- the ISGCI class ID

        - ``recognition_function`` -- a function of one argument `g`, which
          return boolean answers to the question : *does ``g`` belong to the
          class represented by ``gc_id`` ?*

        EXAMPLES::

            sage: graph_classes.Chordal  # indirect doctest
            Chordal graphs
        """
        self._name = name
        self._gc_id = gc_id

        if recognition_function is not None:
            self._recognition_function = recognition_function

    def _repr_(self):
        r"""
        Return a short description of the class.

        EXAMPLES::

            sage: graph_classes.Chordal  # indirect doctest
            Chordal graphs
        """
        return self._name + " graphs"

    def __hash__(self):
        r"""
        Return the class' ID hash.

        EXAMPLES::

            sage: hash(graph_classes.Chordal) == hash(graph_classes.Chordal)
            True
        """
        return hash(self._gc_id)

    def __le__(self, other):
        r"""
        <= operator.

        EXAMPLES::

            sage: graph_classes.Chordal <= graph_classes.Tree
            Unknown
        """
        return other >= self

    def __ge__(self, other):
        r"""
        >= operator.

        EXAMPLES::

            sage: graph_classes.Chordal >= graph_classes.Tree
            True
        """
        inclusion_digraph = GraphClasses().inclusion_digraph()
        if inclusion_digraph.shortest_path(self._gc_id, other._gc_id):
            return True
        else:
            return Unknown

    def __eq__(self, other):
        r"""
        == operator.

        EXAMPLES::

            sage: graph_classes.Chordal == graph_classes.Tree
            Unknown
        """
        return self >= other >= self

    def __lt__(self, other):
        r"""
        >, !=, and < operators.

        EXAMPLES::

            sage: graph_classes.Chordal > graph_classes.Tree
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: graph_classes.Chordal < graph_classes.Tree
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: graph_classes.Chordal != graph_classes.Tree
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    __gt__ = __ne__ = __lt__

    def forbidden_subgraphs(self):
        r"""
        Return the list of forbidden induced subgraphs defining the class.

        If the graph class is not defined by a *finite* list of forbidden
        induced subgraphs, ``None`` is returned instead.

        EXAMPLES::

            sage: graph_classes.Perfect.forbidden_subgraphs()
            sage: gc = graph_classes.get_class('gc_62')
            sage: gc
            claw--free graphs
            sage: gc.forbidden_subgraphs()
            [Graph on 4 vertices]
            sage: gc.forbidden_subgraphs()[0].is_isomorphic(graphs.ClawGraph())
            True
        """
        classes = GraphClasses().classes()
        gc = classes[self._gc_id]

        if gc.get("type", None) != "forbidden":
            return None

        excluded = gc.get("smallgraph", None)

        if not excluded:
            return None

        if not isinstance(excluded, list):
            excluded = [excluded]

        smallgraphs = GraphClasses().smallgraphs()

        if not all(g in smallgraphs for g in excluded):
            return None

        return [smallgraphs[g] for g in excluded]

    def __contains__(self, g):
        r"""
        Check if ``g`` belongs to the graph class represented by ``self``.

        EXAMPLES::

            sage: graphs.CompleteBipartiteGraph(3,3) in graph_classes.Bipartite
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Chordal
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Comparability
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Interval
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Line
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Perfect
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Planar                       # needs planarity
            True
            sage: graphs.CompleteGraph(4) in graph_classes.Split
            True
            sage: graphs.PathGraph(4) in graph_classes.Tree
            True
        """
        from sage.graphs.graph import Graph

        if not isinstance(g, Graph):
            return False

        if hasattr(self, "_recognition_function"):
            return self._recognition_function(g)

        excluded = self.forbidden_subgraphs()

        if excluded is None:
            raise NotImplementedError("No recognition algorithm is available "
                                      "for this class.")

        for gg in excluded:
            if g.subgraph_search(gg, induced=True):
                return False

        return True

    def description(self):
        r"""
        Print the information of ISGCI about the current class.

        EXAMPLES::

            sage: graph_classes.Chordal.description()
            Class of graphs : Chordal
            -------------------------
            id                             :  gc_32
            name                           :  chordal
            ...
            Problems :
            -----------
            3-Colourability                :  Linear
            Clique                         :  Polynomial
            Clique cover                   :  Polynomial
            ...
            Recognition                    :  Linear
            ...
        """
        classes = GraphClasses().classes()
        cls = classes[self._gc_id]

        print("Class of graphs : " + self._name)
        print("-" * (len(self._name) + 18))

        for key, value in sorted(cls.items()):
            if value != "" and key != "problem":
                print("{:30} : {}".format(key, value))

        print("\nProblems :")
        print("-" * 11)

        for pbname, data in sorted(cls["problem"].items()):
            if "complexity" in data:
                print("{:30} : {}".format(pbname, data["complexity"]))


class GraphClasses(UniqueRepresentation):
    def get_class(self, id):
        r"""
        Return the class corresponding to the given id in the ISGCI database.

        INPUT:

        - ``id`` -- string; the desired class' ID

        .. SEEALSO::

            :meth:`~sage.graphs.isgci.GraphClasses.show_all`

        EXAMPLES:

        With an existing id::

            sage: Cographs = graph_classes.get_class("gc_151")
            sage: Cographs
            cograph graphs

        With a wrong id::

            sage: graph_classes.get_class(-1)
            Traceback (most recent call last):
            ...
            ValueError: The given class id does not exist in the ISGCI database. Is the db too old ? You can update it with graph_classes.update_db().
        """
        classes = self.classes()
        if id in classes:
            c = classes[id]

            if c.get("name", ""):
                name = c["name"]
            else:
                name = "class " + str(id)

            return GraphClass(name, id)
        else:
            raise ValueError("The given class id does not exist in the ISGCI "
                             "database. Is the db too old ? You can update it "
                             "with graph_classes.update_db().")

    @cached_method
    def classes(self):
        r"""
        Return the graph classes, as a dictionary.

        Upon the first call, this loads the database from the local XML
        file. Subsequent calls are cached.

        EXAMPLES::

            sage: t = graph_classes.classes()
            sage: type(t)
            <... 'dict'>
            sage: sorted(t["gc_151"].keys())
            ['id', 'name',... 'problem',... 'type']
            sage: t["gc_151"]['name']
            'cograph'
            sage: t["gc_151"]['problem']['Clique']
            {'complexity': 'Linear'}
        """
        self._get_ISGCI()
        return self.classes()

    @cached_method
    def inclusions(self):
        r"""
        Return the graph class inclusions.

        OUTPUT: list of dictionaries

        Upon the first call, this loads the database from the local XML file.
        Subsequent calls are cached.

        EXAMPLES::

            sage: t = graph_classes.inclusions()
            sage: type(t)
            <... 'list'>
            sage: t[0]
            {'sub': 'gc_1', 'super': 'gc_2'}
        """
        self._get_ISGCI()
        return self.inclusions()

    @cached_method
    def smallgraphs(self):
        r"""
        Return a dictionary associating a graph to a graph description string.

        Upon the first call, this loads the database from the local XML files.
        Subsequent calls are cached.

        EXAMPLES::

            sage: t = graph_classes.smallgraphs()
            sage: t['2C_4']
            Graph on 8 vertices
            sage: t['2K_3 + e']
            Graph on 6 vertices
            sage: t['fish']
            Graph on 6 vertices
            sage: t['bull']
            Graph on 5 vertices
        """
        self._get_ISGCI()
        return self.smallgraphs()

    @cached_method
    def inclusion_digraph(self):
        r"""
        Return the class inclusion digraph.

        Upon the first call, this loads the database from the local XML file.
        Subsequent calls are cached.

        EXAMPLES::

            sage: g = graph_classes.inclusion_digraph(); g
            Digraph on ... vertices
        """
        classes = self.classes()
        inclusions = self.inclusions()

        from sage.graphs.digraph import DiGraph
        inclusion_digraph = DiGraph()
        inclusion_digraph.add_vertices(classes.keys())

        for edge in inclusions:
            if edge.get("confidence", "") == "unpublished":
                continue
            inclusion_digraph.add_edge(edge['super'], edge['sub'])

        return inclusion_digraph

    def _download_db(self):
        r"""
        Download the current version of the ISGCI db.

        EXAMPLES::

            sage: graph_classes._download_db()  # optional - internet
        """
        import tempfile
        data_dir = os.path.dirname(DatabaseGraphs().absolute_filename())
        u = urlopen('https://www.graphclasses.org/data.zip',
                    context=default_context())
        with tempfile.NamedTemporaryFile(suffix='.zip') as f:
            f.write(u.read())
            z = zipfile.ZipFile(f.name)

            # Save a systemwide updated copy whenever possible
            try:
                z.extract(_XML_FILE, data_dir)
                z.extract(_SMALLGRAPHS_FILE, data_dir)
            except OSError:
                pass

    def _parse_db(self):
        r"""
        Parse the ISGCI database and stores its content in ``self``.

        EXAMPLES::

            sage: graph_classes._parse_db()
        """
        import xml.etree.ElementTree as ET
        from sage.graphs.graph import Graph

        data_dir = os.path.dirname(DatabaseGraphs().absolute_filename())
        xml_file = os.path.join(data_dir, _XML_FILE)
        tree = ET.ElementTree(file=xml_file)
        root = tree.getroot()
        DB = _XML_to_dict(root)

        classes = {c['id']: c for c in DB['GraphClasses']["GraphClass"]}
        for c in classes.values():
            c["problem"] = {pb.pop("name"): pb for pb in c["problem"]}

        inclusions = DB['Inclusions']['incl']

        # Parses the list of ISGCI small graphs
        smallgraph_file = open(os.path.join(data_dir, _SMALLGRAPHS_FILE))
        smallgraphs = {}

        for line in smallgraph_file.readlines():
            key, string = line.split("\t")
            smallgraphs[key] = Graph(string)

        smallgraph_file.close()

        self.inclusions.set_cache(inclusions)
        self.classes.set_cache(classes)
        self.smallgraphs.set_cache(smallgraphs)

    def update_db(self):
        r"""
        Update the ISGCI database by downloading the latest version from
        internet.

        This method downloads the ISGCI database from the website
        `GraphClasses.org <http://www.graphclasses.org/>`_. It then extracts the
        zip file and parses its XML content. The XML file is saved in the directory
        controlled by the :class:`DatabaseGraphs` class (usually, ``$HOME/.sage/db``).

        EXAMPLES::

            sage: graph_classes.update_db()  # optional - internet
            Database downloaded
        """
        self._download_db()

        print("Database downloaded")

        self.classes.clear_cache()
        self.inclusions.clear_cache()
        self.inclusion_digraph.clear_cache()

    def _get_ISGCI(self):
        r"""
        Return the contents of the ISGCI database.

        This method is mostly for internal use, but often provides useful
        information during debugging operations.

        OUTPUT:

        A pair ``(classes, inclusions)`` where ``classes`` is a dict of dict,
        and ``inclusions`` is a list of dicts.

        .. NOTE::

            This method returns the data contained in the most recent ISGCI
            database present on the computer. See :meth:`update_db` to update
            the latter.

        EXAMPLES::

            sage: graph_classes._get_ISGCI()  # long time (4s on sage.math, 2012)
        """
        self._parse_db()

    def show_all(self):
        r"""
        Print all graph classes stored in ISGCI.

        EXAMPLES::

            sage: graph_classes.show_all()
            id        | name                                     | type                 | smallgraph
            ----------------------------------------------------------------------------------------------------------------------
            gc_309    | $K_4$--minor--free                       | base                 |
            gc_541    | $N^*$                                    | base                 |
            gc_215    | $N^*$--perfect                           | base                 |
            gc_5      | $P_4$--bipartite                         | base                 |
            gc_3      | $P_4$--brittle                           | base                 |
            gc_6      | $P_4$--comparability                     | base                 |
            gc_7      | $P_4$--extendible                        | base                 |
            ...
        """
        classes = self.classes()

        # We want to print the different fields, and this dictionary stores the
        # maximal number of characters of each field.
        MAX = {
            "id": 0,
            "type": 0,
            "smallgraph": 0,
            "name": 0
            }

        # We sort the classes alphabetically, though we would like to display
        # the meaningful classes at the top of the list
        def sort_key(x):
            name = x.get("name", "zzzzz")
            return "{}{:4}".format(name, int(x["id"].split('_')[1]))

        classes_list = sorted(classes.values(), key=sort_key)

        # Maximum width of a field
        MAX_LEN = 40

        # Computing the max of each field with the database
        for key in MAX:
            MAX[key] = len(max((str(x.get(key, "")) for x in classes_list), key=len))

        # At most MAX characters per field
        for key, length in MAX.items():
            MAX[key] = min(length, MAX_LEN)

        # Head of the table
        st = ("{:" + str(MAX["id"]) + "}").format("id")
        st += (" | {:" + str(MAX["name"]) + "}").format("name")
        st += (" | {:" + str(MAX["type"]) + "}").format("type")
        st += (" | {:" + str(MAX["smallgraph"]) + "}").format("smallgraph")
        print(st)
        print("-" * (sum(MAX.values())+9))

        # Entries
        for entry in classes_list:
            ID = entry.get("id", "")
            name = entry.get("name", "")
            typ = entry.get("type", "")
            smallgraph = entry.get("smallgraph", "")
            st = ("{:" + str(MAX["id"]) + "}").format(ID)
            st += (" | {:" + str(MAX["name"]) + "}").format(name[:MAX_LEN])
            st += (" | {:" + str(MAX["type"]) + "}").format(typ[:MAX_LEN])
            st += " | " + str(smallgraph)[:MAX_LEN]
            print(st)


def _XML_to_dict(root):
    r"""
    Return the XML data as a dictionary.

    INPUT:

    - ``root`` -- an ``xml.etree.cElementTree.ElementTree`` object

    OUTPUT: a dictionary representing the XML data

    EXAMPLES::

        sage: graph_classes.Perfect.description()  # indirect doctest
        Class of graphs : Perfect
        -------------------------
        id                             :  gc_56
        name                           :  perfect
        ...
    """
    ans = root.attrib.copy()
    for child in root:
        if child.tag in ans:
            if not isinstance(ans[child.tag], list):
                ans[child.tag] = [ans[child.tag]]
            ans[child.tag].append(_XML_to_dict(child))
        else:
            ans[child.tag] = _XML_to_dict(child)

    # If the dictionary is empty, perhaps the only content is a text, and we
    # return this instead. Useful sometimes in the ISGCI db, for graph names.
    if not ans:
        return root.text
    return ans


graph_classes = GraphClasses()

# Any object added to this list should also appear in the class' documentation, at the top of the file.
graph_classes.Apex = GraphClass("Apex", "gc_1181", recognition_function=lambda x: x.is_apex())
graph_classes.AT_free = GraphClass("AT-free", "gc_61", recognition_function=lambda x: x.is_asteroidal_triple_free())
graph_classes.Biconnected = GraphClass("Biconnected", "gc_771", recognition_function=lambda x: x.is_biconnected())
graph_classes.BinaryTrees = GraphClass("BinaryTrees", "gc_847")
graph_classes.Bipartite = GraphClass("Bipartite", "gc_69", recognition_function=lambda x: x.is_bipartite())
graph_classes.Block = GraphClass("Block", "gc_93", recognition_function=lambda x: x.is_block_graph())
graph_classes.Cactus = GraphClass("Cactus", "gc_108", recognition_function=lambda x: x.is_cactus())
graph_classes.Chordal = GraphClass("Chordal", "gc_32", recognition_function=lambda x: x.is_chordal())
graph_classes.ClawFree = GraphClass("Claw-free", "gc_62")
graph_classes.CoGraph = GraphClass("CoGraph", "gc_151", recognition_function=lambda x: x.is_cograph())
graph_classes.Comparability = GraphClass("Comparability", "gc_72", recognition_function=lambda x: x.is_comparability())
graph_classes.DistanceRegular = GraphClass("Distance Regular", "gc_1148", recognition_function=lambda x: x.is_distance_regular())
graph_classes.Gallai = GraphClass("Gallai", "gc_73")
graph_classes.Grid = GraphClass("Grid", "gc_464")
graph_classes.Hamiltonian = GraphClass("Hamiltonian", "gc_1092", recognition_function=lambda x: x.is_hamiltonian())
graph_classes.Interval = GraphClass("Interval", "gc_234", recognition_function=lambda x: x.is_interval())
graph_classes.Line = GraphClass("Line", "gc_249", recognition_function=lambda x: x.is_line_graph())
graph_classes.Modular = GraphClass("Modular", "gc_50")
graph_classes.Outerplanar = GraphClass("Outerplanar", "gc_110")
graph_classes.Perfect = GraphClass("Perfect", "gc_56", recognition_function=lambda x: x.is_perfect())
graph_classes.Permutation = GraphClass("Permutation", "gc_23", recognition_function=lambda x: x.is_permutation())
graph_classes.Planar = GraphClass("Planar", "gc_43", recognition_function=lambda x: x.is_planar())
graph_classes.Polyhedral = GraphClass("Polyhedral", "gc_986", recognition_function=lambda x: x.is_polyhedral())
graph_classes.Split = GraphClass("Split", "gc_39", recognition_function=lambda x: x.is_split())
graph_classes.StronglyRegular = GraphClass("Strongly Regular", "gc_1185", recognition_function=lambda x: x.is_strongly_regular())
graph_classes.Tree = GraphClass("Tree", "gc_342", recognition_function=lambda x: x.is_tree())
graph_classes.TriangleFree = GraphClass("Triangle Free", "gc_371", recognition_function=lambda x: x.is_triangle_free())
graph_classes.UnitDisk = GraphClass("UnitDisk", "gc_389")
graph_classes.UnitInterval = GraphClass("UnitInterval", "gc_299")
graph_classes.WeaklyChordal = GraphClass("Weakly Chordal", "gc_14", recognition_function=lambda x: x.is_weakly_chordal())



# sage_setup: distribution = sagemath-graphs
r"""
Orientations

This module implements several methods to compute orientations of undirected
graphs subject to specific constraints (e.g., acyclic, strongly connected,
etc.). It also implements some iterators over all these orientations.

**This module contains the following methods**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`orient` | Return an oriented version of `G` according the input function `f`.
    :meth:`orientations` | Return an iterator over orientations of `G`.
    :meth:`acyclic_orientations` | Return an iterator over all acyclic orientations of an undirected graph `G`.
    :meth:`strong_orientation` | Return a strongly connected orientation of the graph `G`.
    :meth:`strong_orientations_iterator` | Return an iterator over all strong orientations of a graph `G`
    :meth:`random_orientation` | Return a random orientation of a graph `G`
    :meth:`minimum_outdegree_orientation` | Return an orientation of `G` with the smallest possible maximum outdegree.
    :meth:`bounded_outdegree_orientation` | Return an orientation of `G` such that every vertex `v` has out-degree less than `b(v)`.
    :meth:`eulerian_orientation` | Return an Eulerian orientation of the graph `G`.

Authors
-------

- Kolja Knauer, Petru Valicov (2017-01-10) -- initial version


Methods
-------
"""
# ****************************************************************************
#       Copyright (C)      2017 Kolja Knauer <kolja.knauer@gmail.com>
#                          2017 Petru Valicov <petru.valicov@lirmm.fr>
#                     2017-2024 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from copy import copy
from sage.graphs.digraph import DiGraph
from sage.graphs.generic_graph import _weight_if_real, _weight_1


def _initialize_digraph(G, edges, name=None, weighted=None, sparse=None,
                        data_structure=None, immutable=None, hash_labels=None):
    r"""
    Helper method to return a directed graph built from ``G``.

    This method returns a digraph with the same set of vertices than the input
    graph ``G`` and with specified edges. The data structure can be
    specified. Furthermore, all attributes of the graph are copied to the
    returned digraph.

    INPUT:

    - ``G`` -- a graph

    - ``edges`` -- iterable; the edges of the digraph to return

    - ``name`` -- string (default: ``None``); the name of the digraph to
      return. By default (``None``), the returned digraph has the same name as
      the input graph ``G``.

    - ``weighted`` -- boolean (default: ``None``); weightedness for the oriented
      digraph. By default (``None``), the graph and its orientation will behave
      the same.

    - ``sparse`` -- boolean (default: ``None``); ``sparse=True`` is an alias for
      ``data_structure="sparse"``, and ``sparse=False`` is an alias for
      ``data_structure="dense"``. Only used when ``data_structure=None``.

    - ``data_structure`` -- string (default: ``None``); one of ``'sparse'``,
      ``'static_sparse'``, or ``'dense'``. See the documentation of
      :class:`DiGraph`.

    - ``immutable`` -- boolean (default: ``None``); whether to create a
      mutable/immutable digraph. Only used when ``data_structure=None``.

      * ``immutable=None`` (default) means that the graph and its orientation
        will behave the same way.

      * ``immutable=True`` is a shortcut for ``data_structure='static_sparse'``

      * ``immutable=False`` means that the created digraph is mutable. When used
        to orient an immutable graph, the data structure used is ``'sparse'``
        unless anything else is specified.

    - ``hash_labels`` -- boolean (default: ``None``); whether to include edge
      labels during hashing of the oriented digraph. This parameter defaults to
      ``True`` if the graph is weighted. This parameter is ignored when
      parameter ``immutable`` is not ``True``. Beware that trying to hash
      unhashable labels will raise an error.

    EXAMPLES::

        sage: from sage.graphs.orientations import _initialize_digraph
        sage: G = Graph([(1, 2)], immutable=True, loops=True, multiedges=True)
        sage: D = _initialize_digraph(G, [])
        sage: D.is_immutable()
        True
        sage: D.allows_loops()
        True
        sage: D.allows_multiple_edges()
        True
        sage: D.edges()
        []
        sage: D.add_edge((2, 3))
        Traceback (most recent call last):
        ...
        ValueError: graph is immutable; please change a copy instead (use function copy())
        sage: G = Graph([(1, 2)])
        sage: D = _initialize_digraph(G, [])
        sage: D.vertices()
        [1, 2]
        sage: D.edges()
        []
        sage: D.add_edge((2, 3)); D.edges()
        [(2, 3, None)]

    TESTS::

        sage: from sage.graphs.orientations import _initialize_digraph
        sage: _initialize_digraph(Graph(), [], data_structure='sparse', immutable=False)
        Traceback (most recent call last):
        ...
        ValueError: you cannot define 'immutable' or 'sparse' when 'data_structure' has a value
        sage: _initialize_digraph(Graph(), [], data_structure='sparse', sparse=True)
        Traceback (most recent call last):
        ...
        ValueError: you cannot define 'immutable' or 'sparse' when 'data_structure' has a value
    """
    # Which data structure should be used ?
    if data_structure is not None:
        # data_structure is already defined so there is nothing left to do
        # here. Did the user try to define too much ?
        if immutable is not None or sparse is not None:
            raise ValueError("you cannot define 'immutable' or 'sparse' "
                             "when 'data_structure' has a value")
    # At this point, data_structure is None.
    elif immutable is True:
        data_structure = 'static_sparse'
        if sparse is False:
            raise ValueError("there is no dense immutable backend at the moment")
    elif immutable is False:
        # If the user requests a mutable digraph and input is immutable, we
        # choose the 'sparse' cgraph backend. Unless the user explicitly
        # asked for something different.
        if G.is_immutable():
            data_structure = 'dense' if sparse is False else 'sparse'
    # At this point, data_structure and immutable are None.
    elif sparse is True:
        data_structure = "sparse"
    elif sparse is False:
        data_structure = "dense"

    if data_structure is None:
        from sage.graphs.base.dense_graph import DenseGraphBackend
        from sage.graphs.base.sparse_graph import SparseGraphBackend
        if isinstance(G._backend, DenseGraphBackend):
            data_structure = "dense"
        elif isinstance(G._backend, SparseGraphBackend):
            data_structure = "sparse"
        else:
            data_structure = "static_sparse"

    if name is None:
        name = G.name()
    if weighted is None:
        weighted = G.weighted()
    if hash_labels is None:
        hash_labels = G._hash_labels

    D = DiGraph(data=[G, edges],
                format='vertices_and_edges',
                data_structure=data_structure,
                multiedges=G.allows_multiple_edges(),
                loops=G.allows_loops(),
                weighted=weighted,
                pos=copy(G.get_pos()),
                name=name,
                hash_labels=hash_labels)

    D.set_vertices(G.get_vertices())

    attributes_to_copy = ('_assoc', '_embedding')
    for attr in attributes_to_copy:
        if hasattr(G, attr):
            copy_attr = {}
            old_attr = getattr(G, attr)
            if isinstance(old_attr, dict):
                for v, value in old_attr.items():
                    try:
                        copy_attr[v] = value.copy()
                    except AttributeError:
                        copy_attr[v] = copy(value)
                setattr(D, attr, copy_attr)
            else:
                setattr(D, attr, copy(old_attr))

    return D


def orient(G, f, weighted=None, data_structure=None, sparse=None,
           immutable=None, hash_labels=None):
    r"""
    Return an oriented version of `G` according the input function `f`.

    INPUT:

    - ``G`` -- an undirected graph

    - ``f`` -- a function that inputs an edge and outputs an orientation of this
      edge

    - ``weighted`` -- boolean (default: ``None``); weightedness for the oriented
      digraph. By default (``None``), the graph and its orientation will behave
      the same.

    - ``sparse`` -- boolean (default: ``None``); ``sparse=True`` is an alias for
      ``data_structure="sparse"``, and ``sparse=False`` is an alias for
      ``data_structure="dense"``. Only used when ``data_structure=None``.

    - ``data_structure`` -- string (default: ``None``); one of ``'sparse'``,
      ``'static_sparse'``, or ``'dense'``. See the documentation of
      :class:`DiGraph`.

    - ``immutable`` -- boolean (default: ``None``); whether to create a
      mutable/immutable digraph. Only used when ``data_structure=None``.

      * ``immutable=None`` (default) means that the graph and its orientation
        will behave the same way.

      * ``immutable=True`` is a shortcut for ``data_structure='static_sparse'``

      * ``immutable=False`` means that the created digraph is mutable. When used
        to orient an immutable graph, the data structure used is ``'sparse'``
        unless anything else is specified.

    - ``hash_labels`` -- boolean (default: ``None``); whether to include edge
      labels during hashing of the oriented digraph. This parameter defaults to
      ``True`` if the graph is weighted. This parameter is ignored when
      parameter ``immutable`` is not ``True``. Beware that trying to hash
      unhashable labels will raise an error.

    OUTPUT: a :class:`DiGraph` object

    .. NOTE::

        This method behaves similarly to method
        :meth:`~sage.graphs.generic_graph.GenericGraph.copy`. That is, the
        returned digraph uses the same data structure by default, unless the
        user asks to use another data structure, and the attributes of the input
        graph are copied.

    EXAMPLES::

        sage: G = graphs.CycleGraph(4); G
        Cycle graph: Graph on 4 vertices
        sage: D = G.orient(lambda e:e if e[0] < e[1] else (e[1], e[0], e[2])); D
        Orientation of Cycle graph: Digraph on 4 vertices
        sage: sorted(D.edges(labels=False))
        [(0, 1), (0, 3), (1, 2), (2, 3)]

    TESTS:

    We make sure that one can get an immutable orientation by providing the
    ``data_structure`` optional argument::

        sage: def foo(e):
        ....:     return e if e[0] < e[1] else (e[1], e[0], e[2])
        sage: G = graphs.CycleGraph(4)
        sage: D = G.orient(foo, data_structure='static_sparse')
        sage: D.is_immutable()
        True
        sage: D = G.orient(foo, immutable=True)
        sage: D.is_immutable()
        True

    Bad input::

        sage: G.orient(foo, data_structure='sparse', sparse=False)
        Traceback (most recent call last):
        ...
        ValueError: you cannot define 'immutable' or 'sparse' when 'data_structure' has a value
        sage: G.orient(foo, data_structure='sparse', immutable=True)
        Traceback (most recent call last):
        ...
        ValueError: you cannot define 'immutable' or 'sparse' when 'data_structure' has a value
        sage: G.orient(foo, immutable=True, sparse=False)
        Traceback (most recent call last):
        ...
        ValueError: there is no dense immutable backend at the moment

    Which backend? ::

        sage: G.orient(foo, data_structure='sparse')._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: G.orient(foo, data_structure='dense')._backend
        <sage.graphs.base.dense_graph.DenseGraphBackend object at ...>
        sage: G.orient(foo, data_structure='static_sparse')._backend
        <sage.graphs.base.static_sparse_backend.StaticSparseBackend object at ...>
        sage: G.orient(foo, immutable=True)._backend
        <sage.graphs.base.static_sparse_backend.StaticSparseBackend object at ...>
        sage: G.orient(foo, immutable=True, sparse=True)._backend
        <sage.graphs.base.static_sparse_backend.StaticSparseBackend object at ...>
        sage: G.orient(foo, immutable=False, sparse=True)._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: G.orient(foo, immutable=False, sparse=False)._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: G.orient(foo, data_structure=None, immutable=None, sparse=True)._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: G.orient(foo, data_structure=None, immutable=None, sparse=False)._backend
        <sage.graphs.base.dense_graph.DenseGraphBackend object at ...>
        sage: G.orient(foo, data_structure=None, immutable=None, sparse=None)._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: H = Graph(data_structure='dense')
        sage: H.orient(foo, data_structure=None, immutable=None, sparse=None)._backend
        <sage.graphs.base.dense_graph.DenseGraphBackend object at ...>
    """
    edges = (f(e) for e in G.edge_iterator())
    name = f"Orientation of {G.name()}"
    return _initialize_digraph(G, edges, name=name, weighted=weighted,
                               data_structure=data_structure, sparse=sparse,
                               immutable=immutable, hash_labels=hash_labels)


def orientations(G, data_structure=None, sparse=None):
    r"""
    Return an iterator over orientations of `G`.

    An *orientation* of an undirected graph is a directed graph such that
    every edge is assigned a direction.  Hence there are `2^s` oriented
    digraphs for a simple graph with `s` edges.

    INPUT:

    - ``G`` -- an undirected graph

    - ``data_structure`` -- one of ``'sparse'``, ``'static_sparse'``, or
      ``'dense'``; see the documentation of :class:`Graph` or :class:`DiGraph`;
      default is the data structure of `G`

    - ``sparse`` -- boolean (default: ``None``); ``sparse=True`` is an alias for
      ``data_structure="sparse"``, and ``sparse=False`` is an alias for
      ``data_structure="dense"``. By default (``None``), guess the most suitable
      data structure.

    .. WARNING::

        This always considers multiple edges of graphs as distinguishable, and
        hence, may have repeated digraphs.

    .. SEEALSO::

        - :meth:`~sage.graphs.graph.Graph.strong_orientation`
        - :meth:`~sage.graphs.orientations.strong_orientations_iterator`
        - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.nauty_directg`
        - :meth:`~sage.graphs.orientations.random_orientation`

    EXAMPLES::

        sage: G = Graph([[1,2,3], [(1, 2, 'a'), (1, 3, 'b')]], format='vertices_and_edges')
        sage: it = G.orientations()
        sage: D = next(it)
        sage: D.edges(sort=True)
        [(1, 2, 'a'), (1, 3, 'b')]
        sage: D = next(it)
        sage: D.edges(sort=True)
        [(1, 2, 'a'), (3, 1, 'b')]

    TESTS::

        sage: G = Graph()
        sage: D = [g for g in G.orientations()]
        sage: len(D)
        1
        sage: D[0]
        Digraph on 0 vertices

        sage: G = Graph(5)
        sage: it = G.orientations()
        sage: D = next(it)
        sage: D.size()
        0

        sage: G = Graph([[1,2,'a'], [1,2,'b']], multiedges=True)
        sage: len(list(G.orientations()))
        4

        sage: G = Graph([[1,2], [1,1]], loops=True)
        sage: len(list(G.orientations()))
        2

        sage: G = Graph([[1,2],[2,3]])
        sage: next(G.orientations())
        Digraph on 3 vertices
        sage: G = graphs.PetersenGraph()
        sage: next(G.orientations())
        An orientation of Petersen graph: Digraph on 10 vertices

    An orientation must have the same ground set of vertices as the original
    graph (:issue:`24366`)::

        sage: G = Graph(1)
        sage: next(G.orientations())
        Digraph on 1 vertex

    Which backend? ::

        sage: next(G.orientations(data_structure='sparse', sparse=True))._backend
        Traceback (most recent call last):
        ...
        ValueError: you cannot define 'immutable' or 'sparse' when 'data_structure' has a value
        sage: next(G.orientations(sparse=True))._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: next(G.orientations(sparse=False))._backend
        <sage.graphs.base.dense_graph.DenseGraphBackend object at ...>
        sage: next(G.orientations())._backend
        <sage.graphs.base.sparse_graph.SparseGraphBackend object at ...>
        sage: G = Graph(1, data_structure='dense')
        sage: next(G.orientations())._backend
        <sage.graphs.base.dense_graph.DenseGraphBackend object at ...>
        sage: G = Graph(1, data_structure='static_sparse')
        sage: next(G.orientations())._backend
        <sage.graphs.base.static_sparse_backend.StaticSparseBackend object at ...>

    Check that the embedding is copied::

        sage: G = Graph([(0, 1), (0, 2), (1, 2)])
        sage: embedding = {0: [1, 2], 1: [2, 0], 2: [0, 1]}
        sage: G.set_embedding(embedding)
        sage: next(G.orientations()).get_embedding() == embedding
        True
        sage: G = Graph()
        sage: G.set_embedding({})
        sage: next(G.orientations()).get_embedding() == {}
        True
    """
    name = G.name()
    if name:
        name = 'An orientation of ' + name

    if not G.size():
        yield _initialize_digraph(G, [], name=name,
                                  data_structure=data_structure, sparse=sparse)
        return

    E = [[(u, v, label), (v, u, label)] if u != v else [(u, v, label)]
         for u, v, label in G.edge_iterator()]
    from itertools import product
    for edges in product(*E):
        yield _initialize_digraph(G, edges, name=name,
                                  data_structure=data_structure, sparse=sparse)


def acyclic_orientations(G):
    r"""
    Return an iterator over all acyclic orientations of an undirected graph `G`.

    ALGORITHM:

    The algorithm is based on [Sq1998]_.
    It presents an efficient algorithm for listing the acyclic orientations of a
    graph. The algorithm is shown to require O(n) time per acyclic orientation
    generated, making it the most efficient known algorithm for generating acyclic
    orientations.

    The function uses a recursive approach to generate acyclic orientations of the
    graph. It reorders the vertices and edges of the graph, creating a new graph
    with updated labels. Then, it iteratively generates acyclic orientations by
    considering subsets of edges and checking whether they form upsets in a
    corresponding poset.

    INPUT:

    - ``G`` -- an undirected graph

    OUTPUT: an iterator over all acyclic orientations of the input graph

    .. NOTE::

        The function assumes that the input graph is undirected and the edges
        are unlabelled.

    EXAMPLES:

    To count the number of acyclic orientations for a graph::

        sage: g = Graph([(0, 3), (0, 4), (3, 4), (1, 3), (1, 2), (2, 3), (2, 4)])
        sage: it = g.acyclic_orientations()
        sage: len(list(it))
        54

    Test for arbitrary vertex labels::

        sage: g_str = Graph([('abc', 'def'), ('ghi', 'def'), ('xyz', 'abc'),
        ....:                ('xyz', 'uvw'), ('uvw', 'abc'), ('uvw', 'ghi')])
        sage: it = g_str.acyclic_orientations()
        sage: len(list(it))
        42

    Check that the method returns properly relabeled acyclic digraphs::

        sage: g = Graph([(0, 1), (1, 2), (2, 3), (3, 0), (0, 2)])
        sage: orientations = set([frozenset(d.edges(labels=false)) for d in g.acyclic_orientations()])
        sage: len(orientations)
        18
        sage: all(d.is_directed_acyclic() for d in g.acyclic_orientations())
        True

    TESTS:

    To count the number of acyclic orientations for a graph with 0 vertices::

        sage: list(Graph().acyclic_orientations())
        []

    To count the number of acyclic orientations for a graph with 1 vertex::

        sage: list(Graph(1).acyclic_orientations())
        []

    To count the number of acyclic orientations for a graph with 2 vertices::

        sage: list(Graph(2).acyclic_orientations())
        []

    Acyclic orientations of a complete graph::

        sage: g = graphs.CompleteGraph(5)
        sage: it = g.acyclic_orientations()
        sage: len(list(it))
        120

    Graph with one edge::

        sage: list(Graph([(0, 1)]).acyclic_orientations())
        [Digraph on 2 vertices, Digraph on 2 vertices]

    Graph with two edges::

        sage: len(list(Graph([(0, 1), (1, 2)]).acyclic_orientations()))
        4

    Cycle graph::

        sage: len(list(Graph([(0, 1), (1, 2), (2, 0)]).acyclic_orientations()))
        6
    """
    if not G.size():
        # A graph without edge cannot be oriented
        return

    from sage.rings.infinity import Infinity
    from sage.combinat.subset import Subsets

    def reorder_vertices(G):
        n = G.order()
        ko = n
        k = n
        G_copy = G.copy()
        vertex_labels = {v: None for v in G_copy.vertices()}

        while G_copy.size() > 0:
            min_val = float('inf')
            uv = None
            for u, v, _ in G_copy.edges():
                du = G_copy.degree(u)
                dv = G_copy.degree(v)
                val = (du + dv) / (du * dv)
                if val < min_val:
                    min_val = val
                    uv = (u, v)

            if uv:
                u, v = uv
                vertex_labels[u] = ko
                vertex_labels[v] = ko - 1
                G_copy.delete_vertex(u)
                G_copy.delete_vertex(v)
                ko -= 2

            if G_copy.size() == 0:
                break

        for vertex, label in vertex_labels.items():
            if label is None:
                vertex_labels[vertex] = ko
                ko -= 1

        return vertex_labels

    def order_edges(G, vertex_labels):
        n = len(vertex_labels)
        m = 1
        edge_labels = {}

        for j in range(2, n + 1):
            for i in range(1, j):
                if G.has_edge(i, j):
                    edge_labels[(i, j)] = m
                    m += 1

        return edge_labels

    def is_upset_of_poset(Poset, subset, keys):
        for (u, v) in subset:
            for (w, x) in keys:
                if (Poset[(u, v), (w, x)] == 1 and (w, x) not in subset):
                    return False
        return True

    def generate_orientations(globO, starting_of_Ek, m, k, keys):
        # Creating a poset
        Poset = {}
        for i in range(starting_of_Ek, m - 1):
            for j in range(starting_of_Ek, m - 1):
                u, v = keys[i]
                w, x = keys[j]
                Poset[(u, v), (w, x)] = 0

        # Create a new graph to determine reachable vertices
        new_G = DiGraph()

        # Process vertices up to starting_of_Ek
        new_G.add_edges([(v, u) if globO[(u, v)] == 1 else (u, v) for u, v in keys[:starting_of_Ek]])

        # Process vertices starting from starting_of_Ek
        new_G.add_vertices([u for u, _ in keys[starting_of_Ek:]] + [v for _, v in keys[starting_of_Ek:]])

        if (globO[(k-1, k)] == 1):
            new_G.add_edge(k, k - 1)
        else:
            new_G.add_edge(k-1, k)

        for i in range(starting_of_Ek, m - 1):
            for j in range(starting_of_Ek, m - 1):
                u, v = keys[i]
                w, x = keys[j]
                # w should be reachable from u and v should be reachable from x
                if w in new_G.depth_first_search(u) and v in new_G.depth_first_search(x):
                    Poset[(u, v), (w, x)] = 1

        # For each subset of the base set of E_k, check if it is an upset or not
        upsets = []
        for subset in Subsets(keys[starting_of_Ek:m-1]):
            if (is_upset_of_poset(Poset, subset, keys[starting_of_Ek:m-1])):
                upsets.append(list(subset))

        for upset in upsets:
            for i in range(starting_of_Ek, m - 1):
                u, v = keys[i]
                if (u, v) in upset:
                    globO[(u, v)] = 1
                else:
                    globO[(u, v)] = 0

            yield globO.copy()

    def helper(G, globO, m, k):
        keys = list(globO.keys())
        keys = keys[0:m]

        if m <= 0:
            yield {}
            return

        starting_of_Ek = 0
        for (u, v) in keys:
            if u >= k - 1 or v >= k - 1:
                break
            else:
                starting_of_Ek += 1

        # s is the size of E_k
        s = m - 1 - starting_of_Ek

        # Recursively generate acyclic orientations
        orientations_G_small = helper(G, globO, starting_of_Ek, k - 2)

        # For each orientation of G_k-2, yield acyclic orientations
        for alpha in orientations_G_small:
            for (u, v) in alpha:
                globO[(u, v)] = alpha[(u, v)]

            # Orienting H_k as 1
            globO[(k-1, k)] = 1
            yield from generate_orientations(globO, starting_of_Ek, m, k, keys)

            # Orienting H_k as 0
            globO[(k-1, k)] = 0
            yield from generate_orientations(globO, starting_of_Ek, m, k, keys)

    # Reorder vertices based on the logic in reorder_vertices function
    vertex_labels = reorder_vertices(G)

    # Create a new graph with updated vertex labels using SageMath, Assuming the graph edges are unlabelled
    new_G = G.relabel(perm=vertex_labels, inplace=False)

    G = new_G

    # Order the edges based on the logic in order_edges function
    edge_labels = order_edges(G, vertex_labels)

    # Create globO array
    globO = {uv: 0 for uv in edge_labels}

    m = len(edge_labels)
    k = len(vertex_labels)
    orientations = helper(G, globO, m, k)

    # Create a mapping between original and new vertex labels
    reverse_vertex_labels = {label: vertex for vertex, label in vertex_labels.items()}

    # Iterate over acyclic orientations and create relabeled graphs
    for orientation in orientations:
        edges = ((u, v) if label else (v, u) for (u, v), label in orientation.items())
        D = _initialize_digraph(G, edges)
        D.relabel(perm=reverse_vertex_labels, inplace=True)
        yield D


def strong_orientation(G):
    r"""
    Return a strongly connected orientation of the graph `G`.

    An orientation of an undirected graph is a digraph obtained by giving an
    unique direction to each of its edges. An orientation is said to be strong
    if there is a directed path between each pair of vertices. See also the
    :wikipedia:`Strongly_connected_component`.

    If the graph is 2-edge-connected, a strongly connected orientation can be
    found in linear time. If the given graph is not 2-connected, the orientation
    returned will ensure that each 2-connected component has a strongly
    connected orientation.

    INPUT:

    - ``G`` -- an undirected graph

    OUTPUT: a digraph representing an orientation of the current graph

    .. NOTE::

        - This method assumes that the input the graph is connected.
        - The time complexity is `O(n+m)` for ``SparseGraph`` and `O(n^2)` for
          ``DenseGraph`` .

    .. SEEALSO::

        - :meth:`~sage.graphs.graph.Graph.orientations`
        - :meth:`~sage.graphs.orientations.strong_orientations_iterator`
        - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.nauty_directg`
        - :meth:`~sage.graphs.orientations.random_orientation`

    EXAMPLES:

    For a 2-regular graph, a strong orientation gives to each vertex an
    out-degree equal to 1::

        sage: g = graphs.CycleGraph(5)
        sage: g.strong_orientation().out_degree()
        [1, 1, 1, 1, 1]

    The Petersen Graph is 2-edge connected. It then has a strongly connected
    orientation::

        sage: g = graphs.PetersenGraph()
        sage: o = g.strong_orientation()
        sage: len(o.strongly_connected_components())
        1

    The same goes for the CubeGraph in any dimension::

        sage: all(len(graphs.CubeGraph(i).strong_orientation().strongly_connected_components()) == 1
        ....:     for i in range(2,6))
        True

    A multigraph also has a strong orientation::

        sage: g = Graph([(0, 1), (0, 2), (1, 2)] * 2, multiedges=True)
        sage: g.strong_orientation()
        Multi-digraph on 3 vertices
    """
    from sage.graphs.base.dense_graph import DenseGraphBackend
    if isinstance(G._backend, DenseGraphBackend):
        data_structure = "dense"
    else:
        data_structure = "sparse"
    d = _initialize_digraph(G, [], data_structure=data_structure)

    i = 0

    # The algorithm works through a depth-first search. Any edge used in the
    # depth-first search is oriented in the direction in which it has been
    # used. All the other edges are oriented backward

    v = next(G.vertex_iterator())
    seen = {}
    i = 1

    # Time at which the vertices have been discovered
    seen[v] = i

    # indicates the stack of edges to explore
    next_ = G.edges_incident(v)

    while next_:
        e = next_.pop()

        # Ignore loops
        if e[0] == e[1]:
            continue

        # We assume e[0] to be a `seen` vertex
        e = e if seen.get(e[0], False) is not False else (e[1], e[0], e[2])

        # If we discovered a new vertex
        if seen.get(e[1], False) is False:
            d.add_edge(e)
            next_.extend(ee for ee in G.edges_incident(e[1])
                         if ((e[0], e[1]) != (ee[0], ee[1])) and ((e[0], e[1]) != (ee[1], ee[0])))
            i += 1
            seen[e[1]] = i

        # Else, we orient the edges backward
        else:
            if seen[e[0]] < seen[e[1]]:
                d.add_edge(e[1], e[0], e[2])
            else:
                d.add_edge(e)

    # Case of multiple edges. If another edge has already been inserted, we add
    # the new one in the opposite direction.
    tmp = None
    for e in G.multiple_edges():
        if tmp == (e[0], e[1]):
            if d.has_edge(e[0], e[1]):
                d.add_edge(e[1], e[0], e[2])
            else:
                d.add_edge(e)
        tmp = (e[0], e[1])

    return d


def strong_orientations_iterator(G):
    r"""
    Return an iterator over all strong orientations of a graph `G`.

    A strong orientation of a graph is an orientation of its edges such that the
    obtained digraph is strongly connected (i.e. there exist a directed path
    between each pair of vertices). According to Robbins' theorem (see the
    :wikipedia:`Robbins_theorem`), the graphs that have strong orientations are
    exactly the 2-edge-connected graphs (i.e., the bridgeless graphs).

    ALGORITHM:

    It is an adaptation of the algorithm published in [CGMRV16]_.
    It runs in `O(mn)` amortized time, where `m` is the number of edges and
    `n` is the number of vertices. The amortized time can be improved to `O(m)`
    with a more involved method.
    In this function, first the graph is preprocessed and a spanning tree is
    generated. Then every orientation of the non-tree edges of the graph can be
    extended to at least one new strong orientation by orienting properly
    the edges of the spanning tree (this property is proved in [CGMRV16]_).
    Therefore, this function generates all partial orientations of the non-tree
    edges and then launches a helper function corresponding to the generation
    algorithm described in [CGMRV16]_.
    In order to avoid trivial symmetries, the orientation of an arbitrary edge
    is fixed before the start of the enumeration process.

    INPUT:

    - ``G`` -- an undirected graph

    OUTPUT: an iterator which will produce all strong orientations of this graph

    .. NOTE::

        Works only for simple graphs (no multiple edges).
        To avoid symmetries an orientation of an arbitrary edge is fixed.

    .. SEEALSO::

        - :meth:`~Graph.orientations`
        - :meth:`~Graph.strong_orientation`
        - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.nauty_directg`
        - :meth:`~sage.graphs.orientations.random_orientation`

    EXAMPLES:

    A cycle has one possible (non-symmetric) strong orientation::

        sage: g = graphs.CycleGraph(4)
        sage: it = g.strong_orientations_iterator()
        sage: len(list(it))
        1

    A tree cannot be strongly oriented::

        sage: g = graphs.RandomTree(10)
        sage: len(list(g.strong_orientations_iterator()))
        0

    Neither can be a disconnected graph::

        sage: g = graphs.CompleteGraph(6)
        sage: g.add_vertex(7)
        sage: len(list(g.strong_orientations_iterator()))
        0

    TESTS:

        sage: g = graphs.CompleteGraph(2)
        sage: len(list(g.strong_orientations_iterator()))
        0

        sage: g = graphs.CubeGraph(3)
        sage: b = True
        sage: for orientedGraph in g.strong_orientations_iterator():
        ....:     if not orientedGraph.is_strongly_connected():
        ....:         b = False
        sage: b
        True

    The total number of strong orientations of a graph can be counted using
    the Tutte polynomial evaluated at points (0,2)::

        sage: g = graphs.PetersenGraph()
        sage: nr1 = len(list(g.strong_orientations_iterator()))
        sage: nr2 = g.tutte_polynomial()(0,2)
        sage: nr1 == nr2/2  # The Tutte polynomial counts also the symmetrical orientations
        True
    """
    # if the graph has a bridge or is disconnected,
    # then it cannot be strongly oriented
    if G.order() < 3 or not G.is_connected() or any(G.bridges(labels=False)):
        return

    V = list(G)

    # compute an arbitrary spanning tree of the undirected graph
    T = G.subgraph(vertices=G, edges=G.min_spanning_tree(), inplace=False)
    treeEdges = list(T.edges(labels=False, sort=False))
    A = [edge for edge in G.edge_iterator(labels=False) if not T.has_edge(edge)]

    # Initialize a digraph with the edges of the spanning tree doubly oriented
    Dg = T.to_directed(sparse=True)
    Dg.add_edges(A)

    # initialization of the first binary word 00...0
    # corresponding to the current orientation of the non-tree edges
    existingAedges = [0] * len(A)

    # Generate all orientations for non-tree edges (using Gray code)
    # Each of these orientations can be extended to a strong orientation
    # of G by orienting properly the tree-edges
    previousWord = 0

    # the orientation of one edge is fixed so we consider one edge less
    nr = 2**(len(A) - 1)
    for i in range(nr):
        word = (i >> 1) ^ i
        bitChanged = word ^ previousWord

        bit = 0
        while bitChanged > 1:
            bitChanged >>= 1
            bit += 1

        previousWord = word
        if not existingAedges[bit]:
            Dg.reverse_edge(A[bit])
            existingAedges[bit] = 1
        else:
            Dg.reverse_edge(A[bit][1], A[bit][0])
            existingAedges[bit] = 0
        # launch the algorithm for enumeration of the solutions
        yield from _strong_orientations_of_a_mixed_graph(Dg, V, treeEdges)


def _strong_orientations_of_a_mixed_graph(Dg, V, E):
    r"""
    Helper function for the generation of all strong orientations.

    Generates all strong orientations of a given partially directed graph
    (also called mixed graph). The algorithm finds bound edges i.e undirected
    edges whose orientation is forced and tries all possible orientations for
    the other edges. See [CGMRV16]_ for more details.

    INPUT:

    - ``Dg`` -- the mixed graph. The undirected edges are doubly oriented

    - ``V`` -- the set of vertices

    - ``E`` -- the set of undirected edges (they are oriented in both ways);
      no labels are allowed

    OUTPUT: an iterator which will produce all strong orientations of the input
    partially directed graph

    EXAMPLES::

        sage: from sage.graphs.orientations import _strong_orientations_of_a_mixed_graph
        sage: g = graphs.CycleGraph(5)
        sage: Dg = DiGraph(g)  # all edges of g will be doubly oriented
        sage: it = _strong_orientations_of_a_mixed_graph(Dg, list(g), list(g.edges(labels=False, sort=False)))
        sage: len(list(it))  # there are two orientations of this multigraph
        2
    """
    length = len(E)
    i = 0
    boundEdges = []
    while i < length:
        u, v = E[i]
        Dg.delete_edge(u, v)
        if v not in Dg.depth_first_search(u):
            # del E[i] in constant time
            E[i] = E[-1]
            E.pop()
            length -= 1
            Dg.add_edge(u, v)
            Dg.delete_edge(v, u)
            boundEdges.append((v, u))
        else:
            Dg.add_edge(u, v)
            Dg.delete_edge(v, u)
            if u not in Dg.depth_first_search(v):
                # del E[i] in constant time
                E[i] = E[-1]
                E.pop()
                length -= 1
                boundEdges.append((u, v))
                Dg.delete_edge(u, v)
            else:
                i += 1
            Dg.add_edge(v, u)

    # if true the obtained orientation is strong
    if not E:
        yield Dg.copy()
    else:
        u, v = E.pop()
        Dg.delete_edge(v, u)
        for orientation in _strong_orientations_of_a_mixed_graph(Dg, V, E):
            yield orientation
        Dg.add_edge(v, u)
        Dg.delete_edge(u, v)
        for orientation in _strong_orientations_of_a_mixed_graph(Dg, V, E):
            yield orientation
        Dg.add_edge(u, v)
        E.append((u, v))
    Dg.add_edges(boundEdges)
    E.extend(boundEdges)


def random_orientation(G):
    r"""
    Return a random orientation of a graph `G`.

    An *orientation* of an undirected graph is a directed graph such that every
    edge is assigned a direction. Hence there are `2^m` oriented digraphs for a
    simple graph with `m` edges.

    INPUT:

    - ``G`` -- a Graph

    EXAMPLES::

        sage: from sage.graphs.orientations import random_orientation
        sage: G = graphs.PetersenGraph()
        sage: D = random_orientation(G)
        sage: D.order() == G.order(), D.size() == G.size()
        (True, True)

    TESTS:

    Giving anything else than a Graph::

        sage: random_orientation([])
        Traceback (most recent call last):
        ...
        ValueError: the input parameter must be a Graph

    .. SEEALSO::

        - :meth:`~Graph.orientations`
        - :meth:`~Graph.strong_orientation`
        - :meth:`~sage.graphs.orientations.strong_orientations_iterator`
        - :meth:`~sage.graphs.digraph_generators.DiGraphGenerators.nauty_directg`
    """
    from sage.graphs.graph import Graph
    if not isinstance(G, Graph):
        raise ValueError("the input parameter must be a Graph")

    from sage.misc.prandom import getrandbits
    rbits = getrandbits(G.size())
    edges = []
    for u, v, l in G.edge_iterator():
        edges.append((u, v, l) if rbits % 2 else (v, u, l))
        rbits >>= 1

    return _initialize_digraph(G, edges, name=f"Random orientation of {G.name()}")


def minimum_outdegree_orientation(G, use_edge_labels=False, solver=None, verbose=0,
                                  *, integrality_tolerance=1e-3):
    r"""
    Return an orientation of `G` with the smallest possible maximum outdegree.

    Given a Graph `G`, it is polynomial to compute an orientation `D` of the
    edges of `G` such that the maximum out-degree in `D` is minimized. This
    problem, though, is NP-complete in the weighted case [AMOZ2006]_.

    INPUT:

    - ``G`` -- an undirected graph

    - ``use_edge_labels`` -- boolean (default: ``False``)

      - When set to ``True``, uses edge labels as weights to compute the
        orientation and assumes a weight of `1` when there is no value available
        for a given edge.

      - When set to ``False`` (default), gives a weight of 1 to all the edges.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set to 0
      by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring;
      see :meth:`MixedIntegerLinearProgram.get_values`.

    EXAMPLES:

    Given a complete bipartite graph `K_{n,m}`, the maximum out-degree of an
    optimal orientation is `\left\lceil \frac {nm} {n+m}\right\rceil`::

        sage: g = graphs.CompleteBipartiteGraph(3,4)
        sage: o = g.minimum_outdegree_orientation()                                     # needs sage.numerical.mip
        sage: max(o.out_degree()) == integer_ceil((4*3)/(3+4))                          # needs sage.numerical.mip
        True

    Show the influence of edge labels on the solution::

        sage: # needs sage.numerical.mip
        sage: g = graphs.PetersenGraph()
        sage: o = g.minimum_outdegree_orientation(use_edge_labels=False)
        sage: max(o.out_degree())
        2
        sage: _ = [g.set_edge_label(u, v, 1) for u, v in g.edge_iterator(labels=False)]
        sage: o = g.minimum_outdegree_orientation(use_edge_labels=True)
        sage: max(o.out_degree())
        2
        sage: g.set_edge_label(0, 1, 100)
        sage: o = g.minimum_outdegree_orientation(use_edge_labels=True)
        sage: max(o.out_degree())
        3

    TESTS::

        sage: from sage.graphs.orientations import minimum_outdegree_orientation
        sage: minimum_outdegree_orientation(DiGraph())                                  # needs sage.numerical.mip
        Traceback (most recent call last):
        ...
        ValueError: Cannot compute an orientation of a DiGraph.
         Please convert it to a Graph if you really mean it.
    """
    G._scream_if_not_simple()
    if G.is_directed():
        raise ValueError("Cannot compute an orientation of a DiGraph. "
                         "Please convert it to a Graph if you really mean it.")

    if use_edge_labels:
        def weight(e):
            label = G.edge_label(e[0], e[1])
            return _weight_if_real(label)
    else:
        weight = _weight_1

    from sage.numerical.mip import MixedIntegerLinearProgram

    p = MixedIntegerLinearProgram(maximization=False, solver=solver)
    degree = p.new_variable(nonnegative=True)

    # The orientation of an edge is boolean and indicates whether the edge uv
    # goes from u to v ( equal to 0 ) or from v to u ( equal to 1)
    orientation = p.new_variable(binary=True)

    # Whether an edge adjacent to a vertex u counts positively or negatively. To
    # do so, we first fix an arbitrary extremity per edge uv.
    ext = {frozenset(e): e[0] for e in G.edge_iterator(labels=False)}

    def outgoing(u, e, variable):
        if u == ext[frozenset(e)]:
            return variable
        return 1 - variable

    for u in G:
        p.add_constraint(p.sum(weight(e) * outgoing(u, e, orientation[frozenset(e)])
                               for e in G.edge_iterator(vertices=[u], labels=False))
                         - degree['max'], max=0)

    p.set_objective(degree['max'])

    p.solve(log=verbose)

    orientation = p.get_values(orientation, convert=bool, tolerance=integrality_tolerance)

    # Return the resulting orientation
    return G.orient(lambda e: e if orientation[frozenset(e[:2])] else (e[1], e[0], e[2]))


def bounded_outdegree_orientation(G, bound, solver=None, verbose=False,
                                  *, integrality_tolerance=1e-3):
    r"""
    Return an orientation of `G` such that every vertex `v` has out-degree less
    than `b(v)`.

    INPUT:

    - ``G`` -- an undirected graph

    - ``bound`` -- maximum bound on the out-degree. Can be of three
      different types :

      * An integer `k`. In this case, computes an orientation whose maximum
        out-degree is less than `k`.

      * A dictionary associating to each vertex its associated maximum
        out-degree.

      * A function associating to each vertex its associated maximum
        out-degree.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set to 0
      by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring;
      see :meth:`MixedIntegerLinearProgram.get_values`.

    OUTPUT:

    A DiGraph representing the orientation if it exists.
    A :exc:`ValueError` exception is raised otherwise.

    ALGORITHM:

    The problem is solved through a maximum flow :

    Given a graph `G`, we create a ``DiGraph`` `D` defined on `E(G)\cup V(G)\cup
    \{s,t\}`. We then link `s` to all of `V(G)` (these edges having a capacity
    equal to the bound associated to each element of `V(G)`), and all the
    elements of `E(G)` to `t` . We then link each `v \in V(G)` to each of its
    incident edges in `G`. A maximum integer flow of value `|E(G)|` corresponds
    to an admissible orientation of `G`. Otherwise, none exists.

    EXAMPLES:

    There is always an orientation of a graph `G` such that a vertex `v` has
    out-degree at most `\lceil \frac {d(v)} 2 \rceil`::

        sage: g = graphs.RandomGNP(40, .4)
        sage: b = lambda v: integer_ceil(g.degree(v)/2)
        sage: D = g.bounded_outdegree_orientation(b)
        sage: all(D.out_degree(v) <= b(v) for v in g)
        True

    Chvatal's graph, being 4-regular, can be oriented in such a way that its
    maximum out-degree is 2::

        sage: g = graphs.ChvatalGraph()
        sage: D = g.bounded_outdegree_orientation(2)
        sage: max(D.out_degree())
        2

    For any graph `G`, it is possible to compute an orientation such that
    the maximum out-degree is at most the maximum average degree of `G`
    divided by 2. Anything less, though, is impossible.

        sage: g = graphs.RandomGNP(40, .4)
        sage: mad = g.maximum_average_degree()                                      # needs sage.numerical.mip

    Hence this is possible ::

        sage: d = g.bounded_outdegree_orientation(integer_ceil(mad/2))              # needs sage.numerical.mip

    While this is not::

        sage: try:                                                                  # needs sage.numerical.mip
        ....:     g.bounded_outdegree_orientation(integer_ceil(mad/2-1))
        ....:     print("Error")
        ....: except ValueError:
        ....:     pass

    The bounds can be specified in different ways::

        sage: g = graphs.PetersenGraph()
        sage: b = lambda v: integer_ceil(g.degree(v)/2)
        sage: D = g.bounded_outdegree_orientation(b)
        sage: b_dict = {u: b(u) for u in g}
        sage: D = g.bounded_outdegree_orientation(b_dict)
        sage: unique_bound = 2
        sage: D = g.bounded_outdegree_orientation(unique_bound)

    TESTS:

    As previously for random graphs, but more intensively::

        sage: for i in range(30):      # long time (up to 6s on sage.math, 2012)
        ....:     g = graphs.RandomGNP(40, .4)
        ....:     b = lambda v: integer_ceil(g.degree(v)/2)
        ....:     D = g.bounded_outdegree_orientation(b)
        ....:     if not (
        ....:          all( D.out_degree(v) <= b(v) for v in g ) or
        ....:          D.size() != g.size()):
        ....:         print("Something wrong happened")

    Empty graph::

        sage: Graph().bounded_outdegree_orientation(b)
        Digraph on 0 vertices
    """
    G._scream_if_not_simple()
    n = G.order()

    if not n:
        return DiGraph()

    vertices = list(G)
    vertices_id = {y: x for x, y in enumerate(vertices)}

    b = {}

    # Checking the input type. We make a dictionary out of it
    if isinstance(bound, dict):
        b = bound
    else:
        try:
            b = dict(zip(vertices, map(bound, vertices)))
        except TypeError:
            b = dict(zip(vertices, [bound]*n))

    d = DiGraph()

    # Adding the edges (s,v) and ((u,v),t)
    d.add_edges(('s', vertices_id[v], b[v]) for v in vertices)

    d.add_edges(((vertices_id[u], vertices_id[v]), 't', 1)
                 for u, v in G.edges(sort=False, labels=None))

    # each v is linked to its incident edges

    for u, v in G.edge_iterator(labels=None):
        u, v = vertices_id[u], vertices_id[v]
        d.add_edge(u, (u, v), 1)
        d.add_edge(v, (u, v), 1)

    # Solving the maximum flow
    value, flow = d.flow('s', 't', value_only=False, integer=True,
                         use_edge_labels=True, solver=solver, verbose=verbose,
                         integrality_tolerance=integrality_tolerance)

    if value != G.size():
        raise ValueError("No orientation exists for the given bound")

    # The flow graph may not contain all the vertices, if they are
    # not part of the flow...
    edges = ((vertices[u], vertices[vv if vv != u else uu])
             for u in (x for x in range(n) if x in flow)
             for uu, vv in flow.neighbors_out(u))

    return _initialize_digraph(G, edges)


def eulerian_orientation(G):
    r"""
    Return an Eulerian orientation of the graph `G`.

    An Eulerian graph being a graph such that any vertex has an even degree, an
    Eulerian orientation of a graph is an orientation of its edges such that
    each vertex `v` verifies `d^+(v)=d^-(v)=d(v)/2`, where `d^+` and `d^-`
    respectively represent the out-degree and the in-degree of a vertex.

    If the graph is not Eulerian, the orientation verifies for any vertex `v`
    that `| d^+(v)-d^-(v) | \leq 1`.

    INPUT:

    - ``G`` -- an undirected graph

    ALGORITHM:

    This algorithm is a random walk through the edges of the graph, which
    orients the edges according to the walk. When a vertex is reached which has
    no non-oriented edge (this vertex must have odd degree), the walk resumes at
    another vertex of odd degree, if any.

    This algorithm has time complexity in `O(n+m)` for ``SparseGraph`` and
    `O(n^2)` for ``DenseGraph``, where `m` is the number of edges in the graph
    and `n` is the number of vertices in the graph.

    EXAMPLES:

    The CubeGraph with parameter 4, which is regular of even degree, has an
    Eulerian orientation such that `d^+ = d^-`::

        sage: g = graphs.CubeGraph(4)
        sage: g.degree()
        [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
        sage: o = g.eulerian_orientation()
        sage: o.in_degree()
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
        sage: o.out_degree()
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]

    Secondly, the Petersen Graph, which is 3 regular has an orientation such
    that the difference between `d^+` and `d^-` is at most 1::

        sage: g = graphs.PetersenGraph()
        sage: o = g.eulerian_orientation()
        sage: o.in_degree()
        [2, 2, 2, 2, 2, 1, 1, 1, 1, 1]
        sage: o.out_degree()
        [1, 1, 1, 1, 1, 2, 2, 2, 2, 2]

    TESTS::

        sage: E0 = Graph(); E4 = Graph(4)  # See trac #21741
        sage: E0.eulerian_orientation()
        Digraph on 0 vertices
        sage: E4.eulerian_orientation()
        Digraph on 4 vertices
    """
    d = _initialize_digraph(G, [], sparse=True, immutable=False)

    if not G.size():
        return d

    g = copy(G)

    # list of vertices of odd degree
    odd = [x for x in g.vertex_iterator() if g.degree(x) % 2]

    # Picks the first vertex, which is preferably an odd one
    if odd:
        v = odd.pop()
    else:
        v = next(g.edge_iterator(labels=None))[0]
        odd.append(v)
    # Stops when there is no edge left
    while True:

        # If there is an edge adjacent to the current one
        if g.degree(v):
            e = next(g.edge_iterator(v))
            g.delete_edge(e)
            if e[0] != v:
                e = (e[1], e[0], e[2])
            d.add_edge(e)
            v = e[1]

        # The current vertex is isolated
        else:
            odd.remove(v)

            # jumps to another odd vertex if possible
            if odd:
                v = odd.pop()
            # Else jumps to an even vertex which is not isolated
            elif g.size():
                v = next(g.edge_iterator())[0]
                odd.append(v)
            # If there is none, we are done !
            else:
                return d
